from __future__ import absolute_import, division, print_function
#from __future__ import  unicode_literals


import sys
import os

from mmgroup.bitfunctions import bitparity, bitweight
from mmgroup.dev.hadamard.hadamard_codegen import HadamardMatrixCode
from mmgroup.generate_c import UserDirective, UserFormat


class HadamardOpT64(HadamardMatrixCode):
    """Code for multiplying vector by a modified Hadamard matrix

    This class supports the operation of the triality element t**e,
    e = 1,2  on a block v with tag T in the representation 196884x 
    of the monster group as described in[Seys19]. We have 
    t**3 == 1. Here t is a 64 times 64 matrix acting on v by right
    multiplication We have t = M * D and t**2 = D * M for a
    diagonal matrix D and a Hadamard-linke matrix M as described 
    below. 

    For D we have D[i,i] = (-1)**((bitweight(i) + 1) >> 1).

    M is the parity-adjusted 2**n times 2**n Hadamard matrix 
    for an even number n as defined in the header of module
    define_matrices.py. This class supports the generation of
    C code for multiplying a vector of integers modulo p
    by a parity-adjusted Hadamard matrix M.

    Notation and implementation of vectors is as in the base 
    class HadamardMatrixCode, which supports right-multiplication 
    of a vector by a Hadamard  matrix.

    To 'upgrade' multiplication by a Hadamard matrix to
    multiplication by a parity-adjusted Hadamard matrix M, we 
    need to code a permutation of the entries of a vector v 
    that exchanges v[i] with v[63 - i] if i has odd parity 
    and that fixes v[i] if i has even parity.
    """
    def __init__(self, p, log_vlength = 6, verbose = 0):
        """Create an instance of the class.

        Parameters are as in the base class HadamardMatrixCode,
        with the restriction that log_vlength must be even.
        """
        args = (p, log_vlength, verbose)
        super(HadamardOpT64, self).__init__(*args)
        assert log_vlength & 1== 0
        self.make_parity_masks()
        self.make_diagonal_mask()
        self.directives.update(self.make_directives())

    def make_parity_masks(self):
        """Create parity masks

        self.PARITY_MASK[0] and self.PARITY_MASK[1] select the 
        entries v[j] of a vector stored in a single variable, 
        where index j has even and odd parity, respectively.
        """
        parity = sum((bitparity(i) << i for i in range(self.INT_FIELDS)))
        mask0 = self.smask(self.P, ~parity)
        mask1 = self.smask(self.P, parity)
        self.PARITY_MASK = [mask0, mask1]

    def make_diagonal_mask(self):
        """Create masks for multiplication with diagonal matrix

        For the diagonal matrix D we have 

           D[i,i] = (-1)**((bitweight(i) + 1) >> 1),   0 <= i < 64,
        
        where bitweight(i) is the bit weight of the binary number i.

        For j, k let i = j * self.INT_FIELDS + k. Then we set
        bit field k in self.DIAG_MASK[j] to 0 if D[i,i] = 1 and to
        self.P if D[i,i] = -1.        
        """
        diag = sum((bool((bitweight(i)+1) & 2) << i for i in range(64)))
        self.DIAG_MASK = []
        for i in range(0, 64, self.INT_FIELDS):
            self.DIAG_MASK.append(self.smask(self.P, diag >> i)) 

    def merge_parities(self, dest, var0, var1, parity = 0):
        """Merge entries of dest from var0 and var1

        Here dest, var0, and var1 are vectors stored in a
        single variable.

        If the integer 'parity' has even bit parity then the
        entry var0[i] copied to dest[i] if i has even parity
        and the entry var1[i] is copied to dest[i] if i has
        odd parity. If the integer 'parity' has odd parity, 
        the roles of var0 and var1 are echanged.
        """
        mask0, mask1 = self.PARITY_MASK
        if bitparity(parity) == 1:
            mask0, mask1 = mask1, mask0
        dest.assign((var0 & mask0) | (var1 & mask1))

    def swap_fields(self, var):
        """Swap entries of vector stored in variable 'var'.

        Here var is a variable storing a vector 
        var[0], ..., var[m - 1], for m = self.INT_FIELD.
        The function exchanges var[i] with var[m - i - 1].
        """
        for i0 in range(self.LOG_INT_FIELDS - 1, -1, -1):
            i = 1 << i0
            msk = self.B0_MASK[i]
            sh = i << self.LOG_FIELD_BITS
            if sh << 1 == self.INT_BITS:
                var.assign((var << sh) | (var >> sh))
            else:
                var.assign(((var & msk) << sh) | ((var >> sh) & msk))

    def swap_odd_parities_log_nvars_even(self):
        """Auxiliary method for method swap_parities.

        It handles the case when self.LOG_INT_FIELDS is odd.
        """
        l = len(self.vars)
        t0, t1 = self.vars.temp(0), self.vars.temp(1)
        for i in range(0, l >> 1, 2):
            v0, v1 = self.vars[i], self.vars[i + 1]
            v2, v3 = self.vars[l - i - 2], self.vars[l - i - 1]
            self.merge_parities(t0, v0, v1, i ^ 1)
            self.swap_fields(t0)
            self.merge_parities(t1, v2, v3, i)
            self.swap_fields(t1)
            self.merge_parities(v0, v0, t1, i)
            self.merge_parities(v1, v1, t1, i ^ 1)
            self.merge_parities(v2, v2, t0, i ^ 1)
            self.merge_parities(v3, v3, t0, i)
            
            
    def swap_odd_parities_log_nvars_odd(self):
        """Auxiliary method for method swap_parities.

        It handles the case when self.LOG_INT_FIELDS is even.
        """
        l = len(self.vars)
        t0 = self.vars.temp(0)
        for i in range(0, l >> 1, 1):
            v0, v1 = self.vars[i], self.vars[l - i - 1]
            self.merge_parities(t0, v0, v1, i ^ 1)
            self.swap_fields(t0)
            self.merge_parities(v0, v0, t0, i)
            self.merge_parities(v1, v1, t0, i ^ 1)
            
    def swap_parities(self):
        """Do a certain permutaition on the entries of the vector

        The function operates in a vector v[0],...,v[2**n-1] 
        stored in variables as described in the base class
        HadamardMatrixCode. It exchanges v[i] with v[2**n-i-1]
        if i has odd parity and it fixes v[i] if i has odd
        parity.
        
        Here n is the parameter log_vlength given in the 
        constructor. The function is defined for even n only.
        """
        if self.LOG_VLEN & 1:
            s = "Size of Hadamard matrix must be a power of 4"
            raise ValueError(s)
        if len(self.vars) < 2:
            s = "Size of Hadamard matrix too small for p = %d"
            raise NotImplementedError(s % self.P)
        odd =  self.LOG_INT_FIELDS & 1
        self.comment(
"""Exchange component i with component %d-i if i 
has odd parity; fix it if i has even parity."""
           % ((1 << self.LOG_VLEN) - 1)
        ) 
        if odd:
            self.swap_odd_parities_log_nvars_odd()
        else: 
            self.swap_odd_parities_log_nvars_even()


    def load_vector_mul_diagonal(self, array_name, mask):
        """Load a vector v form an (external) array A of integers.

        Here 'mask' is an integer of type uint_mmv_t wich must be
        0 or (uint_mmv_t)(-1). If mask == -1 then the vector is 
        multiplied with the dagonal matrix D described in method
        make_diagonal_mask().

        'array_name' must a variable of type uint_mmv_t 
        referring to array A as in method load_vector_direct(). 
        """
        self.comment(
"""Loading vector v from array %s; multiply v
with diagonal matrix if %s == -1."""
            % (array_name, mask)
        )
        for i,v in enumerate(self.vars):
            s = "%s = %s[%d] ^ ((%s) & %s);\n" % (v, 
                array_name, i, mask, self.hex(self.DIAG_MASK[i])) 
            self.add(s, 1, 2)
        self.comment_vector()  

    def store_vector_mul_diagonal(self, mask, array_name):
        """Store a vector v to an (external) array A of integers.

        Here 'mask' is an integer of type uint_mmv_t wich must be
        0 or (uint_mmv_t)(-1). If mask == -1 then the vector is 
        multiplied with the diagonal matrix D described in method
        make_diagonal_mask().

        'array_name' must a variable of type uint_mmv_t 
        referring to array A.
        """ 
        self.comment(
"""Storing vector v to array %s; multiply v
with diagonal matrix if %s == -1."""
            % (array_name, mask)
        )
        s = ""
        for i,v in enumerate(self.vars):
            s = "%s[%d] = %s ^ ((%s) & %s);\n" % (array_name, i, 
                v, mask, self.hex(self.DIAG_MASK[i])) 
            self.add(s, 1, 2)


    def make_code(self, source, mask, dest):
        """Apply triality operation on vector.

        Right multiply the vector of integers mod self.P stored in 
        'src' by t**e, where t is the 64 times 64 triality matrix 
        operating on blocks of the rep 196884x with tag. 

        For e = 1, parameter 'mask' must be 0; for e = 2, parameter
        'mask' must bei (uint_mmv_t)(-1). The result is stored
        in 'dest'.

        'source' and 'dest' must be pointers of type uint_mmv_t*,
        'mask' must be a variable of type uint_mmv_t.
        """
        self.reset_vars()
        self.comment(
"""Multiply the vector of integers mod {p} stored
in ({src}) by t**e, where t is the 64 times 64 
triality matrix and e = 1 if {mask} = 0, e = 2 if
{mask} = (uint_mmv_t)(-1). The result is stored
in ({dest}).
""".format(p = self.P, src = source, mask = mask, dest = dest)
        )
        self.load_vector_mul_diagonal(source, mask)
        self.complement_variable(mask)
        self.swap_parities() 
        self.hadamard_op(shift = -3)
        #self.mul_pwr2(-3)
        self.store_vector_mul_diagonal(mask, dest)
        self.complement_variable(mask)
        self.comment_statistics()
        return self.generate()


    def make_directives(self):
        return {
          "MUL_MATRIX_T64" : UserDirective(self.make_code, "sss"),
        }
        



class HadamardOpT3(HadamardMatrixCode):
    """Apply triality element to tags A, B, C

    Yet to be documented!!!!
    """
        
    def __init__(self, p, verbose = 0):
        """Create an instance of the class.

        Parameters are as in the base class HadamardMatrixCode,
        with the restriction that log_vlength must be even.
        """
        super(HadamardOpT3, self).__init__(p, 0, verbose)
        self.set_vars()
        self.directives.update(self.make_directives())

    def set_vars(self):
        self.vlen = 3 << self.NO_CARRY
        self.vars.resize(self.vlen)
        

    def reset_vars(self):
        super(HadamardOpT3, self).reset_vars()
        self.set_vars()


    def main_op(self):
        """yet to be documented!!!"""
        if self.NO_CARRY:
            for i in range(3):
                self.expand_hadamard_extern(self.vars[i], self.vars[i+3])
            self.external_butterfly(self.vars[4], self.vars[5])
            self.mul_var_pwr2(self.vars[4], -1)
            self.mul_var_pwr2(self.vars[5], -1)
            self.external_butterfly(self.vars[3], self.vars[5])
        self.external_butterfly(self.vars[1], self.vars[2])
        self.mul_var_pwr2(self.vars[1], -1)
        self.mul_var_pwr2(self.vars[2], -1)
        self.external_butterfly(self.vars[0], self.vars[2])
        if self.NO_CARRY:
            for i in range(3):
                self.compress_hadamard_extern(self.vars[i], self.vars[i+3])
        self.vars.xch(0,1)


    def load_vector_mul_diagonal(self, array_name, mask):
        """Load vector v from tags A, B, C of rep 196884x.

        Here 'mask' is an integer of type uint_mmv_t wich must be
        0 or (uint_mmv_t)(-1). If mask == -1 then the vector v is 
        multiplied with the dagonal matrix D described in method
        make_diagonal_mask().

        'array_name' must a variable of type uint_mmv_t 
        referring to array A as in method load_vector_direct(). 
        More specifically, A contains a set of componets of
        a vector for the rep 196884x with tag A.
        The function also loads the corresponding sets of 
        components of the vector with tags B and C.
        """
        self.comment(
"""Loading vector from rep 196884x with tags A,B,C
to v[0...2]. Here %s refers to the tag A part. 
Negate v[2] if %s == -1."""
            % (array_name, mask)
        )
        s = "%s = (%s)[0];\n" % (self.vars[0], array_name)
        s += "%s = (%s)[%d];\n" % (self.vars[1], array_name,
              24 * self.V24_INTS )
        s += "%s = (%s)[%d] ^ ((%s) & %s);\n" % (self.vars[2], 
                array_name, 48 * self.V24_INTS, mask, 
                self.hex(self.smask(self.P)) ) 
        self.add(s, 3, 2)
        self.comment_vector()  

    def store_vector_mul_diagonal(self, mask, array_name):
        """Store vector v to tags A, B, C of rep 196884x.

        Here 'mask' is an integer of type uint_mmv_t wich must be
        0 or (uint_mmv_t)(-1). If mask == -1 then the vector v is 
        multiplied with the dagonal matrix D described in method
        make_diagonal_mask().

        'array_name' must a variable of type uint_mmv_t 
        referring to array A as in method load_vector_direct(). 
        More specifically, A contains a set of componets of
        a vector for the rep 196884x with tag A.
        The function also loads the corresponding sets of 
        components of the vector with tags B and C.
        """
        self.comment(
"""Store vector v[0...2] to rep 196884x with 
tags A,B,C. Here %s refers to the tag A part. 
Negate v[2] if %s == -1."""
            % (array_name, mask)
        )
        s = "(%s)[0] = %s;\n" % (array_name, self.vars[0])
        s += "(%s)[%d] = %s;\n" % (array_name, 
              24 * self.V24_INTS, self.vars[1] )
        s += "(%s)[%d]  = %s ^ ((%s) & %s);\n" % ( 
                array_name, 48 * self.V24_INTS, self.vars[2],
                mask,  self.hex(self.smask(self.P)) ) 
        self.add(s, 3, 2)


    def make_code(self, source, mask, dest):
        """Apply triality operation on vector.

        Right multiply the vector of integers mod self.P stored in 
        'src' by t**e, where t is the 64 times 64 triality matrix 
        operating on blocks of the rep 196884x with tag. 

        For e = 1, parameter 'mask' must be 0; for e = 2, parameter
        'mask' must bei (uint_mmv_t)(-1). The result is stored
        in 'dest'.

        'source' and 'dest' must be pointers of type uint_mmv_t*,
        'mask' must be a variable of type uint_mmv_t.
        """
        self.reset_vars()
        self.comment(
"""Multiply the vector of integers mod {p} stored in
({src}) by t**e, where t is the 3 times 3 triality
matrix [[0, 2,  -2], [1, 1, 1], [1,  -1, -1]] / 2.
and e = 1 if {mask} = 0, e = 2 if {mask} = 
(uint_mmv_t)(-1). The result is stored in ({dest}).

{src} and {dest} are pointers of type *uint_mmv_t.
Components with tags A, B, C referred by ({src}) 
are processed, one integer of type uint_mmv_t
for each tag.

""".format(p = self.P, src = source, mask = mask, dest = dest)
        )
        self.load_vector_mul_diagonal(source, mask)
        self.complement_variable(mask)
        self.main_op()
        self.store_vector_mul_diagonal(mask, dest)
        self.complement_variable(mask)
        self.comment_statistics()
        return self.generate()

    def make_directives(self):
        return {
          "MUL_MATRIX_T3" : UserDirective(self.make_code, "sss"),
        }
        
########################################################################################


class HadamardOpT3A(HadamardMatrixCode):
    """Apply triality element to tags A, B, C

    This is a simplified version of class HadamardOpT3. It
    computes the A part of the vector (multipled by a triality
    element) only.

    Yet to be documented!!!!
    """
        
    def __init__(self, p, verbose = 0):
        """Create an instance of the class.

        Parameters are as in the base class HadamardMatrixCode,
        with the restriction that log_vlength must be even.
        """
        super(HadamardOpT3A, self).__init__(p, 0, verbose)
        self.set_vars()
        self.directives.update(self.make_directives())

    def set_vars(self):
        self.vlen = 2 << self.NO_CARRY
        self.vars.resize(self.vlen)
        

    def reset_vars(self):
        super(HadamardOpT3A, self).reset_vars()
        self.set_vars()


    def external_add(self, v1, v2):
        """Compute  v1 += v2.
   
        Perform addition of the two C variables v1 and v2.  For all 
        components v1[i], v2[i] of of v1, v2
        (stored in v1 and v2) we put

            v1[i] = v1[i] + v2[i], 

        Each result is reduced modulo p. This is a simplified version
        of method external_butterfly().
        """
        if self.FAST_MOD3:
            self.add_mod3(v1, v2, v1)
        else:
            v1.assign(v1 + v2)
            self.reduce_butterfly(v1)

    def main_op(self):
        """yet to be documented!!!"""
        if self.NO_CARRY:
            for i in range(2):
                self.expand_hadamard_extern(self.vars[i], self.vars[i+2])
            self.external_add(self.vars[2], self.vars[3])
        self.external_add(self.vars[0], self.vars[1])
        if self.NO_CARRY:
            self.compress_hadamard_extern(self.vars[0], self.vars[2])
        self.mul_var_pwr2(self.vars[0], -1)




    def load_vector_mul_diagonal(self, array_name, mask):
        """Load vector v from tags A, B, C of rep 196884x.

        Here 'mask' is an integer of type uint_mmv_t wich must be
        0 or (uint_mmv_t)(-1). If mask == -1 then the vector v is 
        multiplied with the dagonal matrix D described in method
        make_diagonal_mask().

        'array_name' must a variable of type uint_mmv_t 
        referring to array A as in method load_vector_direct(). 
        More specifically, A contains a set of componets of
        a vector for the rep 196884x with tag A.
        The function also loads the corresponding sets of 
        components of the vector with tags B and C.
        """
        self.comment(
"""Loading vector from rep 196884x with tags A,B,C
to v[0...2]. Here %s refers to the tag A part. 
Negate v[2] if %s == -1."""
            % (array_name, mask)
        )
        s = "%s = (%s)[%d];\n" % (self.vars[0], array_name,
              24 * self.V24_INTS )
        s += "%s = (%s)[%d] ^ ((%s) & %s);\n" % (self.vars[1], 
                array_name, 48 * self.V24_INTS, mask, 
                self.hex(self.smask(self.P)) ) 
        self.add(s, 2, 2)
        self.comment_vector()  

    def store_vector(self, array_name):
        """Store vector v to tag A, B, C of  of rep 196884x.

        'array_name' must a variable of type uint_mmv_t 
        referring to array A as in method load_vector_direct(). 
        More specifically, A contains a set of componets of
        a vector for the rep 196884x with tag A.
        The function also loads the corresponding sets of 
        components of the vector with tags B and C.
        """
        self.comment(
"""Store vector v[0] to rep 196884x with 
tags A. Here %s refers to the tag A part. """
            % (array_name)
        )
        s = "(%s)[0] = %s;\n" % (array_name, self.vars[0])
        self.add(s, 1)


    def make_code(self, source, mask, dest):
        """Apply triality operation on vector.

        Right multiply the vector of integers mod self.P stored in 
        'src' by t**e, where t is the 64 times 64 triality matrix 
        operating on blocks of the rep 196884x with tag. 

        For e = 1, parameter 'mask' must be 0; for e = 2, parameter
        'mask' must bei (uint_mmv_t)(-1). The part with tag 'A' of
        the result result is stored in 'dest'.

        'source' and 'dest' must be pointers of type uint_mmv_t*,
        'mask' must be a variable of type uint_mmv_t.
        """
        self.reset_vars()
        self.comment(
"""Put dest_A =  (src_B + mask * src_C) / 2   (mod {p})

Here src_B and src_C are the part of a vector of integers 
mod {p} stored in ({src}) with tag B and C, and dest_A is 
the part of a vector of integers mod {p} stored in ({dest}),  
with tag A. Here {mask} must be 0 or -1.

This means that the function computes the part with tag A of
the vector ({dest}) = ({src}) * t**e, where e = 1 - mask.

{src} and {dest} are pointers of type *uint_mmv_t.
Components with tags B, C referred by ({src}) 
are processed, one integer of type uint_mmv_t
for each tag.

""".format(p = self.P, src = source, mask = mask, dest = dest)
        )
        self.load_vector_mul_diagonal(source, mask)
        self.main_op()
        self.store_vector(dest)
        self.comment_statistics()
        return self.generate()

    def make_directives(self):
        return {
          "MUL_MATRIX_T3A" : UserDirective(self.make_code, "sss"),
        }
        







if __name__ == "__main__":
    for p in (3, 7, 127):
        cg = HadamardOpT64(p)
        cg.swap_parities()
        cg.comment_statistics()
        print("\n", cg.generate())



r"""Generation of genernal C code for the monster representations

This script generates the C code for basic computations in the monster
group. The generated C modules are used in the python extension 
``mmgroup.mm``. The functions used by that extension are contained in 
a shared library with name ``mmgroup_mm_basics``. The reason for
creating such a shared library is that that these functions are also
called by other python extensions.

Therefore several .c, .h, .pyd files are created, and also one .pyx
file. Then the Cython package is used for integrating all C 
functions into a single python extension. The action of the extension
is controlled by the .pyx file, which also includes the generated 
.pyd files.

Function ``generate_pyx()`` in this module generates all these files.

Variable ``INT_BITS`` in file config.py should be set to 32 or 64 for
32-bit or 64-bit target systems. There might be performace issues
if ``INT_BITS`` is not chosen properly. 


Generating .c files
...................

We use the C code generation mechanism in class 
``generate_c.TableGenerator``. There .c and .h files are generated
from files with extension .ske. These .ske files are like .c files, 
but augmented with some code generation statements for entering 
(usually rather large) tables and automatically generated code into 
the .c file to be generated. These .ske files may also have statements 
for automatically generating .h files declaring the exported functions.  
    
Variable C_BASICS_SKELETONS is a list of all .ske files. We will 
create  one  .c file from each .ske file in that list.


Entering tables and automatically generated code
................................................

We create an instance ``tg`` of class ``TableGenerator`` for 
generating the .c files. The table generator ``tg`` takes two
dictionaries ``tables`` and ``directives`` as arguments. These
dictionaries provide user-defined tables and directives for the
code generator. There are several table-providing classes that
provide dictionaries ``tables`` and ``directives`` for the 
code genertor ``tg``. Variable ``BASIC_TABLE_CLASSES`` contains 
the list of table-providing classes.

Each table-providing class has methods ``tables()`` and 
``directives()`` that return the corresponding dictionaries.
The union of the dictionaries  ``tables()`` or ``directives()`` 
of all table-providing classes is passed as the constructor
of the instance ``tg`` as argument ``tables()`` or 
``directives()``, respectively.

A simple example for a table-generating class is class 
``Lsbit24Function`` in module ``mmgroups.dev.mat24.mat24aux``.


Generating the .h and the .pxd file
...................................

We create a single .h file containing the prototypes for the 
functions in all generated .c files. The name of that .h file is 
given by the variable ``H_BASICS_NAME``. The content of the .h file 
is also  generated by the instance ``tg`` of class 
``TableGenerator`` described above. 

The .h file has a prefix given by the string ``H_BASICS_BEGIN``
and suffix given by the string ``H_BASICS_END``. 

For each .h file the Cython package requires a .pxd file for
generating a python extension. The .pxd file contains essentially 
the same information as the .h file. Method ``generate_pxd()``
of the code generator ``tg`` generates the .pxd file. The name
of the .pxd file is given by ``PXD_BASICS_NAME``. The generated
.pxd file is prefixed with the content of the string
``PXD_DECLARATIONS``.



Generating the .pxi file
........................

The functions exported by the python extension ``mmgroup.mm`` are
just simple wrappers for the C functions generated by the code
generator. These wrappers must be coded in the Cython language.
They are simple enough so that they can be generated automatically
by function  ``pxd_to_pyx()`` in module ``mmgroup.generate_c``.
Function ``pxd_to_pyx()`` takes the .pxd file (see last section)
as input and it creates the .pxi file with name given by
``PXI_FILE_NAME``. 

When building the ``mmgroup.mm`` extension we simply include that 
.pxi file into the file ``mm_basics.pyx`` (located in directory
``.../mmgroup/dev/mm_basics``). That .pyx file defines the content
of the extension ``mmgroup.mm``.

 
Generating documentation
........................
 
Function ``make_doc()`` in module  ``mmgroup.generate_c`` is used to
generate documentation from some automaticaly generated .c files. 
Here documentation files have names similar to the names of the
corresponding .c files. Usually they just contain a module
documentation and also the commented function prototypes taken from
the .c files. 


Location of the output files
............................

The location of the generated output files is controlled by certain
variables in module config.py. Each of these variables specifies the
name of a directory.

Files with extension .c, .h go to the directory ``C_DIR``. Files with
extension .pxd, .pxi, .pyx go to the directory ``PXD_DIR``. 

"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import re
import collections
import warnings
from numbers import Integral


from config import SRC_DIR, DEV_DIR,  C_DIR, PXD_DIR
from config import REAL_SRC_DIR 
sys.path.append(REAL_SRC_DIR)

from config import INT_BITS
SKE_DIR = os.path.join(DEV_DIR, "mm_basics")

from mmgroup.generate_c import TableGenerator, make_doc, format_item
from mmgroup.generate_c import pxd_to_pyx

import mmgroup.dev.mm_basics
from mmgroup.dev.mm_basics import mm_aux, mm_tables, mm_basics
from mmgroup.dev.mm_basics import mm_tables_xi, mm_crt





VERBOSE = 0

##########################################################################
# Generating .c files
##########################################################################


H_BASICS_SKELETONS = [
    "mm_basics.h",
]

C_BASICS_SKELETONS = [
   "mm_aux",
 #  "mm_sparse",  # deprecated!
 #  "mm_random",  # deprecated!
   "mm_tables",
   "mm_group_word",
   "mm_tables_xi",
   "mm_crt",
]



##########################################################################
# Generating .h and .pxd files files
##########################################################################

H_BASICS_NAME = "mm_basics.h"      # name of generated .h file


PXD_BASICS_NAME = "mm_basics.pxd"  # name of generated .pxd file



### Generate the basic header


H_BASICS_BEGIN = """

// %%EXPORT_KWD MM_BASICS_API


// %%GEN h
#ifndef MM_BASICS_H
#define MM_BASICS_H


"""


H_BASICS_END = """
// %%GEN h

#endif  // #ifndef MM_BASICS_H
"""




### Declarations for the generated .pxd files


PXD_DECLARATIONS = """

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.stdint cimport int64_t, int32_t
from libc.stdint cimport uint{INT_BITS}_t as uint_mmv_t

INT_BITS = {INT_BITS}
""".format(INT_BITS = INT_BITS)




##########################################################################
# Entering tables and automatically generated code
##########################################################################


if "mockup" in sys.argv[1:]:
    BASIC_TABLE_CLASSES = [
        mm_basics.Mockup_MM_Const,
        mm_tables_xi.Mockup_MM_TablesXi,
    ]  
else:
    BASIC_TABLE_CLASSES = [
        mm_basics.MM_Const,
        mm_tables_xi.MM_TablesXi,
    ]  


MORE_BASIC_TABLE_CLASSES = [
    mm_basics.MM_Const,
    mm_aux.MM_IO24,
    mm_tables.MM_OctadTable,
    mm_crt.MM_CrtCombine,  
]

BASIC_TABLE_CLASSES += MORE_BASIC_TABLE_CLASSES

    
##########################################################################
# Generating the .pyx file
##########################################################################



PXI_FILE_NAME = "mm_basics.pxi"







##########################################################################
# Functions of this module for generating .c, .h and .pyx file
##########################################################################

def list_basics_c_files():
    """Return list of names of c files generated by make_basics()"""
    def c_file(name):
         return os.path.join(C_DIR, name + ".c")
    return [c_file(name) for name in C_BASICS_SKELETONS]


def make_basics():
    """Generate basic .c files from  .ske files 

    The relevant .ske files are listed in variable
    C_BASICS_SKELETONS. One corresponding .h and .pxd file
    is also generated. The .c files and the .h file are written 
    to directory given by C_DIR, the .pxd file is written to the 
    current directory.

    Tables and directives for automatically generated C code
    are taken from the classes listed in BASIC_TABLE_CLASSES.

    Return pair of lists, one of the .c files and one of the.
    .pxd files generated
    """
    tables = {}
    directives = {}
    global generated_tables
    for table_class in BASIC_TABLE_CLASSES:
        table_instance = table_class()
        tables.update(table_instance.tables)
        directives.update(table_instance.directives)
    # print("Basic functions:\n",  directives.keys())
    tg = TableGenerator(tables, directives, verbose = VERBOSE)
    # first generate C files
    c_files = []
    all_ske_files = [H_BASICS_BEGIN] 
    all_ske_files += [
        os.path.join(SKE_DIR, f) for f in H_BASICS_SKELETONS]
    for name in C_BASICS_SKELETONS:
        ske_file = name + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        c_file = name + ".c"
        c_path = os.path.join(C_DIR, c_file)
        print("Creating %s from %s" % (c_file, ske_file))
        tg.generate(ske_path, c_path)
        all_ske_files.append(ske_path)
        c_files.append(c_path)
    all_ske_files.append(H_BASICS_END)

    # generate .h and .pxd file
    h_file =  H_BASICS_NAME
    h_path =  os.path.join(C_DIR, h_file)
    pxd_file =  PXD_BASICS_NAME
    print("Creating %s from previous .ske files" % h_file)
    #print(all_ske_files)
    tg.generate(all_ske_files, None, h_path)
    print("Creating %s" % pxd_file)
    tg.generate_pxd(
        os.path.join(PXD_DIR, pxd_file), 
        h_file, 
        PXD_DECLARATIONS
    )
    return c_files,  [ pxd_file ]







##########################################################################
# The main function for generating code
##########################################################################


def generate_files():
    """The main function of this module for generating code

    This function generates all reqired .c files and also the required
    headers.

    It also generates the .pxi file with name given by PYX_FILE_NAME and
    stores it in the current directory. Cython will use the .pxi file to
    build a wrapper for all generated C functions. 
    """
    def pxi_comment(text, f):
        print("\n" + "#"*70 + "\n### %s\n" % text + "#"*70 + "\n\n",file=f)
    c_files,  pxd_files =  make_basics() 
    print("Creating %s" % PXI_FILE_NAME)
    f_pxi = open(os.path.join(PXD_DIR, PXI_FILE_NAME), "wt")
    print(PXD_DECLARATIONS, file = f_pxi)
    for pxd_f in pxd_files:
        pxi_comment("Wrappers for C functions from file %s" % pxd_f, f_pxi)
        pxi_content = pxd_to_pyx(
            os.path.join(PXD_DIR, pxd_f),
            os.path.split(pxd_f)[0]
        )
        print(pxi_content, file = f_pxi)
    f_pxi.close()
    print("Code generation for module mm terminated successfully")
    return c_files


##########################################################################
# Main function
##########################################################################





if __name__ == "__main__":
    generate_files()



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


from random import shuffle, randint
import pytest

from mmgroup.bitfunctions import lrange


from mmgroup.dev.mat24.benes_network import  benes_step
from mmgroup.dev.mat24.benes_network import  check_benes_step
from mmgroup.dev.mat24.benes_network import  build_benes24
from mmgroup.dev.mat24.benes_network import  benes_step_final




def do_benes_step(n, sh, p):
    res0, res1, q = benes_step(n, sh, p)
    check_benes_step(n, sh, p, res0, res1, q)



def exec_benes24(x, net):
    shifts = [1,2,4,8,16,8,4,2,1]
    for  (sh, msk) in zip(shifts, net):
        msk = (x ^ (x >> sh)) & msk
        x ^=  msk ^  (msk << sh) 
    return x
       



def benes24_testdata(ntests=1000):
    for i in range(10):
        p1 = [0,8,16]; shuffle(p1)
        p = [ [p1[j]] + lrange(8*j+1,8*j+8) for j in range(3) ]
        x = randint(0, 2**24-1)
        yield x, sum(p, [])

    for i in range(10):
        x = randint(0, 2**24-1)
        p0 = lrange(0,8); shuffle(p0)
        p1 = lrange(8,16); shuffle(p1)
        p2 = lrange(16,24); shuffle(p2)
        yield x , p0 + p1 + p2

    for i in range(ntests):
        x = randint(0, 2**24-1)
        p0 = lrange(0,24); shuffle(p0)
        yield x, p0
         




@pytest.mark.mat24
def test_benes24_network(ntests = 1000):
    print("24-bit permutation network test ...", end="")
    for x, p in benes24_testdata(ntests):
        net =  build_benes24(p[:])
        y = exec_benes24(x, net)
        for i in range(24):
            assert (x >> i) & 1 == (y >> p[i]) & 1, (i, hex(x), hex(y), p)  
    print(" passed")


@pytest.mark.mat24
def test_benes_steps():
    do_benes_step(4,0, [1,2,0,3])
    do_benes_step(4,1, [1,2,0,3])

    for t in range(10):
        for n in (4,8,16,32):
            for sh in range(5):
                if 1 << sh < n:
                    l = lrange(n)
                    shuffle(l)
                    do_benes_step(n, sh, l)

    for t in range(10):
        l = lrange(32)
        shuffle(l)
        for i in range(4):
            _1, _2, l = benes_step(32, i, l)
        benes_step_final(32, 4, l)
            
            
             



# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint32_t, uint16_t, uint8_t, int32_t
from libc.stdint cimport uint64_t
from libc.stdint cimport uint64_t as uint_mmv_t




######################################################################
### Wrappers for C functions from file mm_op15.pxd
######################################################################


cimport cython
cimport mm_op15

cimport cython

@cython.wraparound(False)
@cython.boundscheck(False)
def op_pi(v_in, delta, pi, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t delta_v_ = delta
    cdef uint32_t pi_v_ = pi
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_pi(&v_in_v_[0], delta_v_, pi_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_delta(v_in, delta, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t delta_v_ = delta
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_delta(&v_in_v_[0], delta_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_pi_tag_A(v, pi):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t pi_v_ = pi
    with nogil:
        mm_op15.mm_op15_pi_tag_A(&v_v_[0], pi_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_copy(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_copy(&mv1_v_[0], &mv2_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare_len(mv1, mv2, len):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t len_v_ = len
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_compare_len(&mv1_v_[0], &mv2_v_[0], len_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_compare(&mv1_v_[0], &mv2_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_vector_add(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    with nogil:
        mm_op15.mm_op15_vector_add(&mv1_v_[0], &mv2_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_scalar_mul(factor, mv1):
    cdef int32_t factor_v_ = factor
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    with nogil:
        mm_op15.mm_op15_scalar_mul(factor_v_, &mv1_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare_mod_q(mv1, mv2, q):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t q_v_ = q
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_compare_mod_q(&mv1_v_[0], &mv2_v_[0], q_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xy(v_in, f, e, eps, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t f_v_ = f
    cdef uint32_t e_v_ = e
    cdef uint32_t eps_v_ = eps
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_xy(&v_in_v_[0], f_v_, e_v_, eps_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_omega(v, d):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t d_v_ = d
    with nogil:
        mm_op15.mm_op15_omega(&v_v_[0], d_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_y_tag_A(v, f):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t f_v_ = f
    with nogil:
        mm_op15.mm_op15_y_tag_A(&v_v_[0], f_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_t(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_t(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_t_A(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_t_A(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xi(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op15.mm_op15_xi(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xi_tag_A(v, exp):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t exp_v_ = exp
    with nogil:
        mm_op15.mm_op15_xi_tag_A(&v_v_[0], exp_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_word(v, g, len_g, e, work):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t len_g_v_ = len_g
    cdef int32_t e_v_ = e
    cdef uint_mmv_t[::1] work_v_ = work
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_word(&v_v_[0], &g_v_[0], len_g_v_, e_v_, &work_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_word_tag_A(v, g, len_g, e):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t len_g_v_ = len_g
    cdef int32_t e_v_ = e
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_word_tag_A(&v_v_[0], &g_v_[0], len_g_v_, e_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_store_order_vector(tags, v0):
    cdef uint32_t[::1] tags_v_ = tags
    cdef uint64_t[::1] v0_v_ = v0
    with nogil:
        mm_op15.mm_op15_store_order_vector(&tags_v_[0], &v0_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_load_order_vector(v0):
    cdef uint64_t[::1] v0_v_ = v0
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_load_order_vector(&v0_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_load_order_tag_vector(tag_vector):
    cdef uint32_t[::1] tag_vector_v_ = tag_vector
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_load_order_tag_vector(&tag_vector_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_order_find_in_Gx0(v, g):
    cdef uint64_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_order_find_in_Gx0(&v_v_[0], &g_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_order_find_in_Qx0(v, g, work):
    cdef uint64_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef uint64_t[::1] work_v_ = work
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_order_find_in_Qx0(&v_v_[0], &g_v_[0], &work_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_order_check_in_Gx0(v, g):
    cdef uint64_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_order_check_in_Gx0(&v_v_[0], &g_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_order_Gx0(g, n, h, o):
    cdef uint32_t[::1] g_v_ = g
    cdef uint32_t n_v_ = n
    cdef uint32_t[::1] h_v_ = h
    cdef uint32_t o_v_ = o
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_order_Gx0(&g_v_[0], n_v_, &h_v_[0], o_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_order(g, n, o):
    cdef uint32_t[::1] g_v_ = g
    cdef uint32_t n_v_ = n
    cdef uint32_t o_v_ = o
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_order(&g_v_[0], n_v_, o_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_axes_find_short(v, p_out, n, y0, y1):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] p_out_v_ = p_out
    cdef uint32_t n_v_ = n
    cdef uint32_t y0_v_ = y0
    cdef uint32_t y1_v_ = y1
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_axes_find_short(&v_v_[0], &p_out_v_[0], n_v_, y0_v_, y1_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_count_short(v, p_out):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] p_out_v_ = p_out
    with nogil:
        mm_op15.mm_op15_count_short(&v_v_[0], &p_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_2A_axis_type(v):
    cdef uint64_t[::1] v_v_ = v
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_2A_axis_type(&v_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_store_axis(v, sign):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t sign_v_ = sign
    with nogil:
        mm_op15.mm_op15_store_axis(&v_v_[0], sign_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_reduce_v_axis(v, r):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] r_v_ = r
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_reduce_v_axis(&v_v_[0], &r_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_reduce_axis(a, n, r):
    cdef uint32_t[::1] a_v_ = a
    cdef uint32_t n_v_ = n
    cdef uint32_t[::1] r_v_ = r
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_reduce_axis(&a_v_[0], n_v_, &r_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_reduce_v_baby_axis(v, r):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] r_v_ = r
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_reduce_v_baby_axis(&v_v_[0], &r_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_reduce_G_x0(a, n, r):
    cdef uint32_t[::1] a_v_ = a
    cdef uint32_t n_v_ = n
    cdef uint32_t[::1] r_v_ = r
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_reduce_G_x0(&a_v_[0], n_v_, &r_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_reduce_M(a, n, r):
    cdef uint32_t[::1] a_v_ = a
    cdef uint32_t n_v_ = n
    cdef uint32_t[::1] r_v_ = r
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_reduce_M(&a_v_[0], n_v_, &r_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_eval_A_odd_mod15_aux(v, b, res):
    cdef uint64_t[::1] v_v_ = v
    cdef uint64_t b_v_ = b
    cdef uint64_t[::1] res_v_ = res
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_eval_A_odd_mod15_aux(&v_v_[0], b_v_, &res_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_eval_A(v, v2):
    cdef uint64_t[::1] v_v_ = v
    cdef uint32_t v2_v_ = v2
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op15.mm_op15_eval_A(&v_v_[0], v2_v_)
    return ret_


######################################################################
### Constants
######################################################################


MMV_ENTRIES = 247488

INT_BITS = 64

LOG_INT_BITS = 6

P = 15

FIELD_BITS = 4

LOG_FIELD_BITS = 2

INT_FIELDS = 16

LOG_INT_FIELDS = 4

P_BITS = 4

MMV_INTS = 15468




import subprocess
import os


def compile_gcc(c_names, exe_name):
    args = ["gcc" ] + c_names + ["-o", exe_name]
    subprocess.check_output(args)

def compile_msvc(c_names, exe_name):
    args = ["cl" ] + c_names + ["/Fe" + exe_name]
    subprocess.check_output(args)

def compile_testprogramm(c_names, exe_name):
    """Compile a list of C programs to a single executable program
 
    ``c_names`` is a list of file names of C programs.
    ``exe_name`` is the file name of the executable program.
    """
    if os.name in ["nt"]:
        try:
            compile_msvc(c_names, exe_name)
            return
        except FileNotFoundError:
            pass
    compile_gcc(c_names, exe_name)




import sys
import os
import numpy as np
from numbers import Integral
from random import randint, choice



import pytest

from mmgroup.tests.groups.mgroup_n import MGroupN
from mmgroup.tests.groups.group_n import GroupN

ref_group = MGroupN()
group = GroupN()



def str_group(g):
    return str(list(map(hex, g.data)))

####################################################################
# Test cases
####################################################################


def make_rand_word(tags, length):   
    return [(choice(tags), 'r') for i in range(length)]

def make_testcases():
    for i in range(5):
        for g1 in "dpxyt":
            for g2 in "dpxyt":
                yield [(g1,'r')], [(g2,'r')]
    for i in range(150):
        g1 = make_rand_word("dpxyt", 10)
        g2 = make_rand_word("dpxyt", 10)
        yield g1, g2


####################################################################
# Test group multiplication
####################################################################

def ll_mul_test_case(g1, g2, verbose = 0):
    g1 = group(g1)
    g2 = group(g2)
    g1ref = ref_group(g1) 
    g2ref = ref_group(g2) 
    g12ref = g1ref * g2ref
    g12 = g1 * g2
    assert group(g12ref) == g12, (g12ref, g12)
    assert ref_group(g12) == g12ref
    g12mul = g1.copy().mul(g2)
    assert g12 == g12mul
    g12mul_atoms = g1.copy().mul_atoms(g2)
    assert g12 == g12mul_atoms
    # next test mixed operation
    #assert g12 == g1 * g2ref
    #assert g12ref == g1ref * g2


@pytest.mark.auto_group
def test_ll_mul(verbose = 0):
    print("testing low-level multiplication in group N")
    for g1, g2 in make_testcases():
        ll_mul_test_case(g1, g2, verbose)
    print("passed")



####################################################################
# Test group division 
####################################################################

def ll_div_test_case(g1, g2, verbose = 0):
    g1 = group(g1)
    g2 = group(g2)
    g1ref = ref_group(g1) 
    g2ref = ref_group(g2) 
    g12ref = g1ref / g2ref
    g12 = g1 / g2
    assert group(g12ref) == g12, (g12ref, g12)
    assert ref_group(g12) == g12ref
    g12div = g1.copy().div(g2)
    assert g12 == g12div
    g12div_atoms = g1.copy().div_atoms(g2)
    assert g12 == g12div_atoms
    # next test mixed operation
    #assert g12 == g1 / g2ref
    #assert g12ref == g1ref / g2


@pytest.mark.auto_group
def test_ll_div(verbose = 0):
    print("testing low-level multiplication in group N")
    for g1, g2 in make_testcases():
        ll_div_test_case(g1, g2, verbose)
    print("passed")


####################################################################
# Test coset decomposition
####################################################################

@pytest.mark.auto_group
def test_coset_decomp(verbose = 0):
    for i in  range(50):
        g = group(make_rand_word("dpxyt", 10))
        g_Nx0, g_t = g.right_coset_N_x0()
        assert g == g_Nx0 * g_t
        assert g_Nx0.data[0] == 0
        d = g_t.data 
        assert d[1] | d[2] | d[3] | d[4] == 0
        






import sys
import os
import warnings

from random import randint
from collections import OrderedDict

import numpy as np
import pytest
from operator import __xor__
from functools import reduce

from mmgroup import MM0, MMSpace, MMV
from mmgroup.generators import gen_leech3_op_vector_word
from mmgroup.generators import gen_leech2_scalprod
from mmgroup.clifford12 import bitmatrix64_echelon_h
from mmgroup.clifford12 import leech2_matrix_basis
from mmgroup.clifford12 import leech2_matrix_orthogonal
from mmgroup.clifford12 import bitmatrix64_xch_bits
from mmgroup.clifford12 import bitmatrix64_rot_bits
from mmgroup.clifford12 import bitmatrix64_cap_h
from mmgroup.clifford12 import leech2_matrix_radical
from mmgroup.clifford12 import leech2_matrix_expand



def leech2_matrix_basis_testcases():
    for i in range(3, 500, 17):
        l = [randint(0, 0xffffff) for j in range(i)]
        yield np.array(l, dtype = np.uint32)
   

def leech_echelon(a):
    a = np.copy(a)
    bitmatrix64_xch_bits(a, 12, 0x800);
    bitmatrix64_rot_bits(a, 1, 12,0);
    k=bitmatrix64_echelon_h(a, len(a), 24, 24);
    bitmatrix64_rot_bits(a, 11, 12,0);
    bitmatrix64_xch_bits(a, 12, 0x800);
    return a[:k]


@pytest.mark.qstate
def test_leech2_matrix_basis():
    for a in leech2_matrix_basis_testcases():
        a1 = np.array(a, dtype = np.uint64)
        k1 = bitmatrix64_echelon_h(a1, len(a1), 24, 24)
        a1 = leech_echelon(a1)
        a2 = np.zeros(24, dtype = np.uint64)
        k2 = leech2_matrix_basis(a, len(a), a2, 24)
        assert k1 == k2 # == k2a
        assert (a1[:k1] == a2[:k1]).all()



def leech2_matrix_orthogonal_testcases():
    for i in [1, 3, 8, 11, 12, 15, 19, 23, 23, 24]:
        l = [randint(0, 0xffffff) for j in range(i)]
        yield np.array(l, dtype = np.uint64)
   

@pytest.mark.qstate
def test_leech2_matrix_orthogonal():
    for a in leech2_matrix_orthogonal_testcases():
        a1 = np.copy(a)
        o1 = np.zeros(24, dtype = np.uint64)
        ko1 = leech2_matrix_orthogonal(a, o1, len(a))
        o1 = o1[ko1:]
        ka1 = bitmatrix64_echelon_h(a1, len(a1), 24, 24)
        assert ka1 ==  ko1
        for v in a1:
            for w in o1:
                assert gen_leech2_scalprod(v, w) == 0
                 



@pytest.mark.qstate
def test_leech2_matrix_radical():
    for a in leech2_matrix_orthogonal_testcases():
        a1 = np.copy(a)
        k1 = bitmatrix64_echelon_h(a1, len(a1), 24, 24)
        a1 = a1[:k1]
        o1 = np.zeros(24, dtype = np.uint64)
        ko1 = leech2_matrix_orthogonal(a1, o1, len(a))
        o1 = o1[ko1:]
        ko1 = 24 - ko1
        ko1 = bitmatrix64_echelon_h(o1, len(o1), 24, 24)
        kcap =  bitmatrix64_cap_h(a1, o1, 24, 24);
        #print(k1, ko1, kcap)
        kcap1 = kcap[0]
        radical = a1[kcap1:k1]
        radical = leech_echelon(radical)

        a_copy = np.array(a, dtype = np.uint32)
        a2 = np.zeros(24, dtype = np.uint64)
        k2 = leech2_matrix_radical(a_copy, len(a_copy), a2, 24)
        a2 = a2[:k2]

        assert len(radical) == k2
        assert (a2 == radical).all()
        

@pytest.mark.qstate
def test_leech2_matrix_radical():
    for a in leech2_matrix_orthogonal_testcases():
        k = len(a)
        if (k > 14):
            continue
        v = np.zeros(1 << k, dtype = np.uint32)
        leech2_matrix_expand(a, k, v)
        for j in range(5):
            index = randint(0, (1 << k) - 1)
            terms = [int(x) for i, x in enumerate(a) 
                       if (1 << (k - 1 - i)) & index]
            ref = reduce(__xor__, terms, 0)
            assert v[index] == ref
        
 


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import numpy as np
from random import randint
from importlib import import_module
import time

from mmgroup import MM0, MMV, AutPL
from mmgroup.mm_space import characteristics 
from mmgroup.mm import mm_group_prepare_op_ABC


import pytest

OP_ABC = {}
for p in characteristics():
    mm = import_module('mmgroup.mm%d' % p)
    OP_ABC[p] = mm.op_word_ABC



#######################################################################
# Test function mm_group_prepare_op_ABC
#######################################################################

def perm(src, dest, d = 0):
    return MM0(AutPL(d, zip(list(src),list(dest)), 0))


def op_word_ABC_testcases(ntests = 10):
    test_values = [
        (3, [('A',2,3)], perm([2,3],[0,1])),
        (3, [('A',2,3)], [('l', 1)]),
    ]
    g_once = [
        [],
        [('x','n'),('d', 0x837)],
        [('y','n')],
        [('t','n')],
        [('p','n')],
        [('l','n')],
    ] 
    g_more = [
        [('r','N_x0')],
        [('r','G_x0')],
        [('r','G_x0'), ('r','N_0')],
    ] 
    for p, vs, gs in test_values:
        yield MMV(p)(vs), MM0(gs)
    for g_values, n in zip([g_once, g_more], [1, ntests]):
        for i in range(n):
            for p in characteristics():
                v = MMV(p)('R')
                for g in g_values:
                    yield v, MM0(g)


def display_mm_group_prepare_op_ABC(g):
    a1 = np.zeros(12, dtype = np.uint32)
    len_a1 = mm_group_prepare_op_ABC(g.mmdata, len(g.mmdata), a1)
    g1 = MM0('a', a1[:len_a1])
    print("mm_group_prepare_op_ABC(g)=")
    print(g1)
    if g1 != g:
        print(" This difffers from g!!")
    return g1


def one_test_op_word_ABC(v, g, verbose = 0):
    p = v.p
    if verbose:
        print("Test function mm_op%d_word_ABC, g =" % p)
        print(" ", g)
    w_ref = v * g
    w = MMV(p)()
    res = OP_ABC[p](v.data, g.mmdata, len(g.mmdata), w.data)
    ok = res >= 0
    for tag in "ABC":
        ok = ok and not (w[tag] != w_ref[tag]).any()
        if not ok:
            print("Test function mm_op%d_word_ABC, g =" % p)
            print(" ", g)
            print("mm%d_op_word_ABC returns %d" % (p, res))
            print("Arrays differ in tag %s, diff =" % tag)
            delta = w[tag] - w_ref[tag]
            print(delta)
            display_mm_group_prepare_op_ABC(g)
            if res != 0:
                err = "Function mm_op%d_word_ABC returns %d" % (p, res)
            else:
                err = "Error in function mm_op%d_word_ABC" % p
            raise ValueError(err)
    

@pytest.mark.mm_op
def test_op_word_ABC(verbose = 0):
    print("Testing function mm_op<p>_word_ABC")
    for v, g in op_word_ABC_testcases():
        one_test_op_word_ABC(v, g, verbose = verbose)
    print("Test passed")


#######################################################################
# Benchmark for function mm_group_prepare_op_ABC
#######################################################################


def make_benchmark_sample(n):
    a = np.zeros((n,6), dtype = np.uint32)
    for i in range(n):
        m = MM0('c', 'r').mmdata
        np.copyto(a[i,:len(m)], m)
    return a

def benchmark_mm_op15_word_ABC(a):
    f = OP_ABC[15]
    v = MMV(15)('R').data
    w = MMV(15)().data
    t_start = time.process_time()
    for g in a:
        f(v, g, 6, w)
    t = time.process_time() - t_start
    return t / len(a)
    
def benchmark_mm_op15_word(a):
    from mmgroup.mm15 import op_word as f
    v = MMV(15)('R').data
    w = MMV(15)().data
    t_start = time.process_time()
    for g in a:
        f(v, g, 6, 1, w)
    t = time.process_time() - t_start
    return t / len(a)


def benchmark_mm_op15_map_t(a):
    from collections import defaultdict
    from mmgroup.mm_reduce import mm_reduce_op_2A_axis_type
    d = defaultdict(int)
    axes = []
    for i in range(64):
        axes.append( (MMV(15)('I', 2, 3) * MM0('r', 5)).data )
    mask = len(axes) - 1
    assert mask & (mask + 1) == 0
    w = MMV(15)().data
    op_t_A = MMV(15)().ops.op_t_A
    w1 = np.zeros(24*4, dtype = np.uint64)
    t_start = time.process_time()
    mode = 0x16
    for i, g in enumerate(a):
        axtypes = mm_reduce_op_2A_axis_type(axes[i & mask], g, 6, mode)
        d[(axtypes >> 8) & 0xff] += 1
        d[(axtypes >> 16) & 0xff] += 1
    t = time.process_time() - t_start
    #print(dict(d))
    return t / len(a)



def benchmark_empty(a):
    from mmgroup.mm15 import op_word as f
    v = MMV(15)().data
    w = MMV(15)().data
    t_start = time.process_time()
    for g in a:
        pass
    t = time.process_time() - t_start
    return t / len(a)



@pytest.mark.slow
@pytest.mark.bench
@pytest.mark.mm_op
def test_benchmark_op_word():
    n = 2000
    a = make_benchmark_sample(n)
    t_0 =  benchmark_empty(a)
    t_f =  benchmark_mm_op15_word(a)
    t_fABC =  benchmark_mm_op15_word_ABC(a)
    t_ax =  benchmark_mm_op15_map_t(a)
    print("Average run times:")
    names = ["mm_op15_word", 
       "mm_reduce_op_2A_axis_type",
       "mm_op15_word_ABC",
       "empty loop",
    ]
    for name, value in zip(names, [t_f, t_ax, t_fABC, t_0]):
        print(" Function %-25s: %7.4f ms" % (name, 1000 * value))






    




from mmgroup.dev.mat24 import Mat24

basis = Mat24.basis[12:]


def print_vector(v, final =  0):    
    for j in range(24):
        print((b >> j) & 1, end = "")
        if j %4 == 3: print(" & ", end = " ")
    if not final: print(r"\\")
    else: print("")

print(r"\text{\scriptsize $\begin{matrix}")
for i, b in enumerate(basis):
    print_vector(b, final = i == len(basis) - 1)
print(r" \end{matrix}$}")



from generate_functions import UserDirective
from make_c_tables import TableGenerator 


class TestGen:
    def add_to(x, expr):
         return "{x} = {x} + {expr};\n".format(x = x, expr = str(expr))  

    functions = {
        "AddTo": UserDirective(add_to, "si"),
    }

    tables = {
        "y" : [1,2,3],
        "TABLE_SIZE": 32,
        "mult": 17,
        "pair_list": list(zip(range(3), [7,4,9])),
    }

def generate():
    T = TestGen
    tg = TableGenerator(T.tables, T.functions, verbose = 1)
    tg.generate("testcode.ske", "test.c", "test.h")
    print("Table in test code:")
    print( tg.names )


if __name__ == "__main__":
    generate()

    









from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices, shuffle, sample
from numbers import Integral
from collections import defaultdict

import numpy as np
import pytest

from mmgroup import XLeech2, Xsp2_Co1, PLoop, GCode, AutPL, Cocode, GcVector

from mmgroup import mat24
from mmgroup.generators import gen_leech2_type
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_op_atom
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_start_type4
from mmgroup.generators import gen_leech2_start_type24
from mmgroup.generators import gen_leech2_reduce_type2
from mmgroup.generators import gen_leech2_reduce_type2_ortho
from mmgroup.generators import gen_leech2_reduce_type4

from mmgroup.tests.test_gen_xi.test_leech2_type import TYPE_DATA
from mmgroup.tests.test_gen_xi.test_leech2_type import rand_n_elem
from mmgroup.tests.test_gen_xi.test_leech2_type import xs_vector

# Standard vector in the Leech lattice mod 2 in Leech lattice encoding
# The standard fram \Omega
OMEGA = 0x800000
# The standard type-2 vector \beta
BETA = 0x200 

assert Cocode(BETA).syndrome() == GcVector(0xC)

#####################################################################
# Auxiliary functions
#####################################################################

def xi_reduce_odd_type4(v, verbose = 0):
    r"""Compute power of :math:`\xi` that reduces a vector ``v``

    Let ``v`` be a vector in the Leech lattice mod 2 in Leech
    lattice encoding. We assume that ``v`` is of subtype 0x43.

    We compute an exponent ``e`` such that :math:`\xi^e` maps 
    ``v`` to a vector of subtype 0x42 or 0x44.

    The function returns ``e`` if ``v`` is mapped to type 0x42
    and ``0x100 + e`` if ``v`` is mapped to type 0x44.  A negative
    return value indicates that no such exponent ``e`` exists.
    """
    assert v & 0x800         # Error if cocode part ov v is even
    coc = (v ^ mat24.ploop_theta(v >> 12)) & 0xfff
    # Obtain cocode as table of bit fields of 5 bits
    tab = mat24.syndrome_table(coc & 0x7ff)
    # Check if the syndrome bits are in 3 different MOG columns.
    # We first XOR bit field i with bit field (i-1)(mod 3)
    # and then zero the lowest two bits of each bit field. 
    tab ^= ((tab >> 5) & 0x3ff) ^ ((tab & 0x1f) << 10);
    tab &= 0x739c;
    # Now all three bit fields are nonzero iff the syndrome bits
    # are in three differnt columns. Next add 32 - 4 to each bit
    # field in order to produce a carry if the field is nonzero.
    tab += 0x739c;
    # Next we isolate the three carry bits 
    tab &= 0x8420;
    # Return -1 if all carry bits are set, i.e all syndrome bits
    # are in different columns.
    if (tab == 0x8420):
         return -1
    # Let scalar be the scalar product of the Golay part of v
    # with the standard tetrad \omega
    scalar = (v >> 22) & 1;
    # Exponent for element \xi of G_x0 is 2 - scalar
    exp = 2 - scalar;
    if verbose:
        w = gen_leech2_op_atom(v, 0x60000000 + exp) 
        print("Reducing c = %s, subtype %s, t=%s, e=%d, to v = %s, subtype %s"
            %  (hex(mat24.cocode_syndrome(coc,0)), hex(gen_leech2_subtype(v)),
            hex(tab), exp,  hex(mat24.gcode_to_vect(w >> 12)),
            hex(gen_leech2_subtype(w))
            ))
    # Return exponent for \xi in the lower 4 bits of the retrun value;
    # Return 0 in bit 8 if all syndrome bits of v are in the same
    # MOG column and 1 in bit 8 otherwise.
    return ((tab != 0) << 8) + exp

def xi_reduce_odd_type2(v, verbose = 0):
    r"""Compute power of :math:`\xi` that reduces a vector ``v``

    Let ``v`` be a vector in the Leech lattice mod 2 in Leech
    lattice encoding. We assume that ``v`` is of subtype 0x21.

    We type to compute an exponent ``e`` such that :math:`\xi^e` 
    maps ``v`` to a vector of subtype 0x22.

    The function returns ``e`` if such an eponent exists.  A negative
    return value indicates that no such exponent ``e`` exists.
    """
    assert v & 0x800         # Error if cocode part ov v is even
    # Let scalar be the scalar product of the Golay part of v
    # with the standard tetrad \omega
    scalar = (v >> 22) & 1;
    # Exponent for element \xi of G_x0 is 2 - scalar
    exp = 2 - scalar;
    return exp


def xi_reduce_octad(v, verbose = 0):
    r"""Compute power of :math:`\xi` that reduces a vector ``v``

    Let ``v`` be a vector in the Leech lattice mod 2 in Leech
    lattice encoding. We assume that ``v`` is of subtype 
    0x22, 0x42, or 0x44.

    We try to compute an exponent ``e`` such that :math:`\xi^e`  
    maps ``v`` to a vector of subtype 0x40 or 0x20.

    The function returns ``e`` if such an eponent exists.  A negative
    return value indicates that no such exponent ``e`` exists.

    Assume :math:`v_2 = \lambda_d + \lambda_\delta + \lambda_\epsilon`, 
    where :math:`d \in \mathcal{C}, \delta, \epsilon \in \mathcal{C}^*`, 
    with :math:`d, \delta` grey, even,  :math:`\epsilon` coloured. 
    The function returns
  
    :math:`e=0` if :math:`d=0 \pmod{\Omega}`,
     
    :math:`e=1` if :math:`\delta=\gamma(d) \pmod{\omega} `,

    :math:`e=2` if :math:`\delta=0  \pmod{\omega} `. 
  
    In all other cases there is no suitable exponent :math:`e`.
    """
    #print("Std12", hex(v))
    if v & 0x7ff800 == 0:
        return 0
    if v & 0x7f080f == 0: 
        return 1
    parity = -((0x6996 >> (v & 0xf)) & 1)
    v ^= ((v >> 12) ^ parity) & 0xf
    if v & 0x7f080f == 0: 
        return 2
    return -1 # no exponent found


def xi_reduce_dodecad(v, verbose = 0):
    r"""Compute power of :math:`\xi` that reduces a vector ``v``

    Let ``v`` be a vector in the Leech lattice mod 2 in Leech
    lattice encoding. We assume that ``v`` is of subtype 0x46.

    We compute an exponent ``e`` such that :math:`\xi^e` maps 
    ``v`` to a vector of subtype 0x44.

    The function returns ``e`` if such an eponent exists.  A negative
    return value indicates that no such exponent ``e`` exists.
    """
    # Let ``vect`` be the Golay code part of v as a bit vector.
    vect = mat24.gcode_to_vect(v >> 12)
    # Set bit 4*i of s if all bits 4*i, 4*i+1, 4*i+2, 4*i+3 of 
    # ``vect`` are equal, otherwise clear bit 4*i, for 0 <= i < 6.
    s1 = vect | (vect >> 2); s1 = s1 | (s1 >> 1);
    s0 = vect & (vect >> 2); s0 = s0 & (s0 >> 1); 
    s = (s0 | ~s1) & 0x111111;
    # If the Golay code part of v is a docecad then either no or two
    # bits in s are set. Fail if no bit in s is set.
    if (s == 0):
        return -1
    # Here two bits of s (in two different MOG columns) are set. 
    # Set all bits in a MOG column if one bit is set in that column.
    # Thus the bits being set in s form a grey even octad.
    s *= 15;
    # Let 'coc' be the cocode part of v
    coc = v ^ mat24.ploop_theta(v >> 12)
    # Compute scalar product of octad s and ``coc`` in ``scalar``
    tab = mat24.syndrome_table((mat24.recip_basis[0] ^ coc) & 0x7ff)
    scalar = s ^ (s >> (tab & 31)) ^  (s >> ((tab >> 5) & 31))
    scalar ^=  (s >> ((tab >> 10) & 31))
    scalar &= 1
    # The requested exponent is now equal to ``2 - scalar``.
    exp = 2 - scalar
    return exp

def apply_perm(v, src, dest, n, log_list, verbose = 0):
    r"""Apply permutation to vector in Leech lattice mod 2.
  
    The function computes a permutation :math:`\pi` that maps
    the entries of the array ``src`` of length ``n`` to
    the entries of the array ``dest`` (of the same length) in
    the given order. 

    Let :math:`v_2` be the vector in the Leech lattice mod  2 given 
    by parameter ``v2``. The function returns :math:`v_2 x_\pi`.
    Parameter ``v2`` and the return value are given in Leech
    lattice encoding.
  
    Parameter ``p_res`` points to an integer where the function 
    stores the element :math:`x_\pi` as a generator of the
    monster group as as described  in file ``mmgroup_generators.h``.
    That generator is stored with tag  ``MMGROUP_ATOM_TAG_IP`` so
    that we can compute the inverse of :math:`\pi` very 
    efficiently. 

    We compute the inverse of the lowest permutation (in lexical
    order) that maps ``dest[:n]`` to ``src[:n]``.
    """
    res, p =  mat24.perm_from_map(dest[:n], src[:n])
    assert res > 0, (res, dest[:n], src[:n])
    p_inv =  mat24.inv_perm(p)
    p_num =  mat24.perm_to_m24num(p)
    log_list.append(0xA0000000 + p_num)
    xd = (v >> 12) & 0xfff
    xdelta = (v ^ mat24.ploop_theta(xd)) & 0xfff
    m =  mat24.perm_to_matrix(p_inv)
    xd = mat24.op_gcode_matrix(xd, m)
    xdelta = mat24.op_cocode_perm(xdelta, p_inv)
    v_out = (xd << 12) ^ xdelta ^ mat24.ploop_theta(xd)
    if verbose:
        print("Apply permutation (mapping v to gcode %s):\n%s" %
          (hex(mat24.gcode_to_vect(v_out >> 12)), p_inv)
        )
    return v_out


#####################################################################
# Find and apply permutations in Mat24
#####################################################################

OCTAD = [0,1,2,3,4, 8,9]

def find_octad_permutation(v, result, verbose = 0):
    coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff
    w = mat24.gcode_weight(v >> 12) 
    vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
    src = mat24.vect_to_list(vect, 5)
    syn =  mat24.cocode_syndrome(coc, src[0]) & ~vect
    if syn:
        v5 = (1 << src[0]) | (1 << src[1]) | (1 << src[2])
        v5 |= syn 
        special = mat24.syndrome(v5, 24);
        src = src[:3]
        src.append(mat24.lsbit24(special & vect))
        src.append(mat24.lsbit24(vect & ~(special | v5)))
        src.append(mat24.lsbit24(syn))  
        syn &= ~(1 << src[-1])
        src.append(mat24.lsbit24(syn))  
    return apply_perm(v, src, OCTAD, len(src), result, verbose)


#####################################################################
# Function reduce_type_2
#####################################################################

def reduce_type2(v, verbose = 1):
    r"""Map (orthgonal) short vector in Leech lattice to standard vector

    This is a python implementation of the C function
    ``gen_leech2_reduce_type2`` in file ``gen_leech.c``.
   
    Let ``v \in \Lambda / 2 \Lambda`` of type 2 be given by 
    parameter ``v`` in Leech lattice encoding. 

    Let ``beta`` be the short vector in the Leech  lattice propotional
    to  ``e_2 - e_3``, where ``e_i`` is  the ``i``-th basis vector
    of ``\{0,1\}^{24}``.
   
    Then the function constructs a ``g \in G_{x0}`` 
    that maps ``v`` to ``beta``.
 
    The element ``g`` is returned as a word in the generators
    of ``G_{x0}`` of length ``n \leq 6``. Each atom of the 
    word ``g`` is encoded as  defined in the header 
    file ``mmgroup_generators.h``. 

    The function stores ``g`` as a word of generators in the
    array ``pg_out`` and returns the length  ``n``  of that
    word. It returns a negative number in case of failure, 
    e.g. if ``v`` is not of type 2.
    """
    vtype = gen_leech2_subtype(v)
    if (vtype >> 4) != 2:
        raise ValueError("Vector is not short")
    result = []
    for _i in range(4):
        if verbose:
            coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff
            vt = gen_leech2_subtype(v)
            coc_anchor = 0
            if vt in [0x22]:
                w = mat24.gcode_weight(v >> 12) 
                vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
                coc_anchor = mat24.lsbit24(vect)
            coc_syn = Cocode(coc).syndrome_list(coc_anchor)
            gcode = mat24.gcode_to_vect(v >> 12)
            print("Round %d, v = %s, subtype %s, gcode %s, cocode %s" % 
                (_i, hex(v & 0xffffff), hex(vt), hex(gcode),  coc_syn)
            )
        assert vtype == gen_leech2_subtype(v)
        if vtype == 0x21:
            exp = xi_reduce_odd_type2(v)
            vtype = 0x22
        elif vtype == 0x22:
            exp = xi_reduce_octad(v)            
            if exp < 0:
                w = mat24.gcode_weight(v >> 12) 
                vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
                src = mat24.vect_to_list(vect, 4) 
                dest = [0,1,2,3]
                v = apply_perm(v, src, dest, 4, result, verbose)
                exp = xi_reduce_octad(v) 
                assert exp >= 0
            vtype = 0x20
        elif vtype == 0x20:
            exp =  0
            # map v to stadard cocode word [2,3]
            if v & 0x7fffff != 0x200:
                 syn = (mat24.cocode_syndrome(v, 0))
                 src = mat24.vect_to_list(syn, 2)
                 v  = apply_perm(v, src, [2,3], 2, result, verbose)
            # correct v2 if v2 is the cocode word [2,3] + Omega
            if v & 0x800000:
                atom = 0xC0000200  
                   # operation y_d such that d has odd scalar
                   # product with cocode word [2,3]
                v = gen_leech2_op_atom(v, atom)
                result.append(atom)
            assert v & 0xffffff == 0x200
            return np.array(result, dtype = np.uint32)
        else:
            raise ValueError("WTF")
        if exp: 
            exp = 0xE0000003 - exp
            v = gen_leech2_op_atom(v, exp)
            result.append(exp)
    raise ValueError("WTF1")

#####################################################################
# Test function reduce_type_2
#####################################################################


#####################################################################
# Test data for function reduce_type_2()


def type2_testdata(ntests):
    """The function yields type-2 vectors in the Leech lattice mod 2

    These type-2 vectors are given in **Leech lattice encoding**.
    """
    testdata = [
        0x200,
        0x1000200,
        0x800200,
        0x1800200,
    ]
    for t in testdata:
         yield t
    for ploop, cocode, subtype in TYPE_DATA:
        if subtype in [0x20, 0x22, 0x21]:
            v = xs_vector(ploop, cocode)
            yield v
            for i in range(ntests):
                g =  rand_n_elem()
                yield g.xsp_conjugate(v)

    for i in range(ntests):
        yield XLeech2('r', 2).ord

#####################################################################
# Testing function reduce_type_2


@pytest.mark.gen_xi
def test_reduce_type_2(ntests = 500, verbose = 0):
    """Test function ``reduce_type2`` """
    for n, v in enumerate(type2_testdata(ntests)):
        if verbose:
            print(" \nTest %d, v = %s, subtype  =%s" % 
                (n+1, hex(v), hex(gen_leech2_subtype(v)))
            )
        op = reduce_type2(v, verbose)
        w = gen_leech2_op_word(v, op, len(op))
        assert w & 0xffffff == BETA, hex(w)
        a = np.zeros(6, dtype = np.uint32)
        l = gen_leech2_reduce_type2(v, a)
        if l < 0:
            err = "Error %s in function gen_leech2_reduce_type2"
            raise ValueError(err % hex(l & 0xffffffff))
        a = a[:l]
        assert list(a) == list(op), ((a), (op))  

#####################################################################
# Function reduce_type_2_ortho
#####################################################################

def reduce_type2_ortho(v, verbose = 0):
    r"""Map (orthgonal) short vector in Leech lattice to standard vector

    This is a python implementation of the C function
    ``gen_leech2_reduce_type2_ortho`` in file ``gen_leech.c``.
   
    Let ``v \in \Lambda / 2 \Lambda`` of type 2 be given by 
    parameter ``v`` in Leech lattice encoding. 

    In the real Leech lattice, (the origin of) the vector ``v`` must
    be orthogonal to the standard short vector ``beta``. Here ``beta``
    is the short vector in the Leech  lattice  propotional
    to  ``e_2 - e_3``, where ``e_i`` is  the ``i``-th basis vector
    of ``\{0,1\}^{24}``.
   
    Let ``beta'`` be the short vector in the Leech lattice propotional
    to  ``e_2 + e_3``.  Then the function constructs a ``g \in G_{x0}`` 
    that maps ``v`` to ``beta'`` and fixes ``beta``.
 
    The element ``g`` is returned as a word in the generators
    of ``G_{x0}`` of length ``n \leq 6``. Each atom of the 
    word ``g`` is encoded as  defined in the header 
    file ``mmgroup_generators.h``. 

    The function stores ``g`` as a word of generators in the
    array ``pg_out`` and returns the length  ``n``  of that
    word. It returns a negative number in case of failure, 
    e.g. if ``v`` is not of type 2,  or not orthogonal 
    to ``beta'`` in the real Leech lattice.
    """
    vtype = gen_leech2_subtype(v)
    if (vtype >> 4) != 2:
        raise ValueError("Vector is not short")
    if gen_leech2_type(v ^ 0x200) != 4:
        raise ValueError("Vector not orthogonal to standard vector")
    result = []
    for _i in range(4):
        if verbose:
            coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff
            vt = gen_leech2_subtype(v)
            coc_anchor = 0
            if vt in [0x22]:
                w = mat24.gcode_weight(v >> 12) 
                vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
                coc_anchor = mat24.lsbit24(vect)
            coc_syn = Cocode(coc).syndrome_list(coc_anchor)
            gcode = mat24.gcode_to_vect(v >> 12)
            print("Round %d, v = %s, subtype %s, gcode %s, cocode %s" % 
                (_i, hex(v & 0xffffff), hex(vt), hex(gcode),  coc_syn)
            )
        assert vtype == gen_leech2_subtype(v)
        if vtype == 0x21:
            exp = xi_reduce_odd_type2(v)
            vtype = 0x22
        elif vtype == 0x22:
            exp = xi_reduce_octad(v)            
            if exp < 0:
                w = mat24.gcode_weight(v >> 12) 
                vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
                if vect & 0x0c:
                    vect &= ~0x0c
                    src = mat24.vect_to_list(vect, 2) + [2,3]
                    dest = [0,1,2,3]
                else:
                    src = [2,3] + mat24.vect_to_list(vect, 3)
                    v5 = (1 << src[2]) | (1 << src[3]) | (1 << src[4])
                    v5 |= 0x0c
                    special = mat24.syndrome(v5, 24);
                    src.append(mat24.lsbit24(special & vect))  
                    dest = [2,3,4,5,6,7]
                v = apply_perm(v, src, dest, len(src), result, verbose)
                exp = xi_reduce_octad(v) 
                assert exp >= 0
            vtype = 0x20
        elif vtype == 0x20:
            if ((v & 0xffffff) == 0x800200):
                return np.array(result, dtype = np.uint32)
            syn = (mat24.cocode_syndrome(v, 0)) & ~0xc
            if syn and syn != 3:
                src =  mat24.vect_to_list(syn, 2) + [2, 3]
                v = apply_perm(v, src, [0, 1, 2, 3], 4, result, verbose)
            exp = 2 - ((v >> 23) & 1);
        else:
            raise ValueError("WTF")
        if exp: 
            exp = 0xE0000003 - exp
            v = gen_leech2_op_atom(v, exp)
            result.append(exp)
    raise ValueError("WTF1")


#####################################################################
# Test function reduce_type_2_ortho
#####################################################################




def type2_ortho_testdata(ntests):
    """Yields special type-2 vectors in the Leech lattice mod 2

    These type-2 vectors are orthogonal to the vector BETA in the 
    real Leech lattice and given in **Leech lattice encoding**.
    """
    for v in type2_testdata(2 * ntests):
        if gen_leech2_type(v ^ BETA) == 4:
            yield v



@pytest.mark.gen_xi
def test_reduce_type_2_ortho(ntests = 500, verbose = 0):
    """Test function ``reduce_type2_ortho`` """
    for n, v in enumerate(type2_ortho_testdata(ntests)):
        if verbose:
            print(" \nTest %d, v = %s, subtype  =%s" % 
                (n+1, hex(v), hex(gen_leech2_subtype(v)))
            )
        op = reduce_type2_ortho(v, verbose)
        w = gen_leech2_op_word(v, op, len(op)) 
        assert w & 0xffffff == 0x800200, hex(w)
        b = gen_leech2_op_word(BETA, op, len(op)) 
        assert b == BETA
        a = np.zeros(6, dtype = np.uint32)
        l = gen_leech2_reduce_type2_ortho(v, a)
        if l < 0:
            err = "Error %s in function gen_leech2_reduce_type2_ortho"
            raise ValueError(err % hex(l & 0xffffffff))
        a = a[:l]
        assert list(a) == list(op), ((a), (op))  
    
    

#####################################################################
# Function reduce_type_4_std
#####################################################################

LSTD = [0,1,2,3,4,5]

def reduce_type4_std(v, verbose = 0):
    r"""Map type-4 vector in Leech lattice to standard vector

    This is (almost) a python implementation of the C function
    ``gen_leech2_reduce_type4`` in file ``gen_leech.c``.
   
    Let ``v \in \Lambda / 2 \Lambda`` of type 4 be given by 
    parameter ``v`` in Leech lattice encoding. 

    Let ``Omega`` be the type- vector in the Leech  lattice 
    corresponding to the standard coordinate frame in the Leech
    lattice.
   
    Then the function constructs a ``g \in G_{x0}`` 
    that maps ``v`` to ``Omega``.
 
    The element ``g`` is returned as a word in the generators
    of ``G_{x0}`` of length ``n \leq 6``. Each atom of the 
    word ``g`` is encoded as  defined in the header 
    file ``mmgroup_generators.h``. 

    The function stores ``g`` as a word of generators in the
    array ``pg_out`` and returns the length  ``n``  of that
    word. It returns a negative number in case of failure, 
    e.g. if ``v`` is not of type 4.

    We remark that the C function ``gen_leech2_reduce_type4`` 
    treats certain type-4  vectors ``v`` in a special way
    as indicated in function ``reduce_type4``.
    """
    if verbose:
        print("Transforming  type-4 vector %s to Omega" %
            hex(v & 0x1ffffff) )
    vtype = gen_leech2_subtype(v)
    result = []
    for _i in range(5):
        coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff
        if verbose:
            vt = gen_leech2_subtype(v)
            coc_anchor = 0
            if vt in [0x42, 0x44]:
                w = mat24.gcode_weight(v >> 12) 
                vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
                coc_anchor = mat24.lsbit24(vect)
            coc_syn = Cocode(coc).syndrome_list(coc_anchor)
            gcode = mat24.gcode_to_vect(v >> 12)
            print("Round %d, v = %s, subtype %s, gcode %s, cocode %s" % 
                (_i, hex(v & 0xffffff), hex(vt), hex(gcode),  coc_syn)
            )
        assert vtype == gen_leech2_subtype(v)
        if vtype == 0x48:
            if verbose:
                 res = list(map(hex, result))
                 print("Transformation is\n", res)
            return np.array(result, dtype = np.uint32)
        elif vtype == 0x40:
            if v & 0x7ffbff:
                syn = mat24.cocode_syndrome(coc, 0)
                src = mat24.vect_to_list(syn, 4)
                v = apply_perm(v, src, LSTD, 4, result, verbose)
                #print("after type 40", hex(v),  Cocode(v).syndrome(0))
            exp = 2 - ((v >> 23) & 1)
            vtype = 0x48
        elif vtype in [0x42, 0x44]:
            exp = xi_reduce_octad(v)            
            if exp < 0:
                v = find_octad_permutation(v, result, verbose)
                exp = xi_reduce_octad(v) 
                assert exp >= 0
            vtype = 0x40
        elif vtype == 0x46:
            exp = xi_reduce_dodecad(v, verbose)
            if exp < 0:
                vect = mat24.gcode_to_vect(v >> 12)
                src = mat24.vect_to_list(vect, 4)
                v = apply_perm(v, src, LSTD, len(src), result, verbose)
                exp = xi_reduce_dodecad(v, verbose) 
                assert exp >= 0
            vtype = 0x44
        elif vtype == 0x43:
            exp = xi_reduce_odd_type4(v, verbose)
            if exp < 0:
                vect = mat24.gcode_to_vect(v >> 12)
                syn = mat24.cocode_syndrome(coc, 24)
                src = mat24.vect_to_list(syn, 3)
                #print("coc list", src)
                v = apply_perm(v, src, LSTD[1:], len(src), result, verbose)
                exp = xi_reduce_odd_type4(v, verbose)
                assert exp > 0
            vtype = 0x42 + ((exp & 0x100) >> 7)
            exp &= 3
        else:
            raise ValueError("WTF")
        if exp: 
            exp = 0xE0000003 - exp
            v_old = v
            v = gen_leech2_op_atom(v, exp)
            assert v & 0xfe000000 == 0, (hex(v_old), hex(exp), hex(v))
            result.append(exp)
    raise ValueError("WTF1")


#####################################################################
# Test function reduce_type_4_std
#####################################################################


#####################################################################
# Test data for function reduce_type_4_std)


def type4_testdata(ntests):
    """The function yields type-4 vectors in the Leech lattice mod 2

    These type-4 vectors are given in **Leech lattice encoding**.
    """
    testdata = [
        0x800000,
        0x1800000,
    ]
    for t in testdata:
         yield t
    for ploop, cocode, subtype in TYPE_DATA:
        if subtype in [0x40, 0x42, 0x44, 0x46, 0x43]:
            v = xs_vector(ploop, cocode)
            yield v
            for i in range(ntests):
                g =  rand_n_elem()
                yield g.xsp_conjugate(v)

    for i in range(ntests):
        yield XLeech2('r', 4).ord

    for v in type2_ortho_testdata(ntests):
        yield v ^ BETA

#####################################################################
# Testing function  data for function reduce_type_4_std()

@pytest.mark.gen_xi
def test_reduce_type_4_std(ntests = 500, verbose = 0):
    """Test function ``reduce_type4_std`` """
    for n, v in enumerate(type4_testdata(ntests)):
        if verbose:
            print(" \nTest %d, v = %s, subtype  =%s" % 
                (n+1, hex(v), hex(gen_leech2_subtype(v)))
            )
        op = reduce_type4_std(v, verbose)
        w = gen_leech2_op_word(v, op, len(op)) 
        assert w & 0xffffff == 0x800000, hex(w)


#####################################################################
# Test function reduce_type_4
#####################################################################




def reduce_type4(v, verbose = 0):
    r"""Map type-4 vector in Leech lattice to standard vector

    This is  a python implementation of the C function
    ``gen_leech2_reduce_type4`` in file ``gen_leech.c``.
   
    Let ``v \in \Lambda / 2 \Lambda`` of type 4 be given by 
    parameter ``v`` in Leech lattice encoding. 

    Let ``Omega`` be the type- vector in the Leech  lattice 
    corresponding to the standard coordinate frame in the Leech
    lattice. Let ``beta`` be the short vector in the Leech  
    lattice propotional  to  ``e_2 - e_3``, where ``e_i`` is  
    the ``i``-th basis vector  of ``\{0,1\}^{24}``.
   
    Then the function constructs a ``g \in G_{x0}`` 
    that maps ``v`` to ``Omega``. If ``v + beta`` is of type 2
    and orthogonal to ``beta`` in the real Leech lattice then the 
    returned element ``g`` also fixes ``beta``.
 
    The element ``g`` is returned as a word in the generators
    of ``G_{x0}`` of length ``n \leq 6``. Each atom of the 
    word ``g`` is encoded as  defined in the header 
    file ``mmgroup_generators.h``. 

    The function stores ``g`` as a word of generators in the
    array ``pg_out`` and returns the length  ``n``  of that
    word. It returns a negative number in case of failure, 
    e.g. if ``v`` is not of type 4.
    """
    vtype = gen_leech2_subtype(v)
    if (vtype >> 4) != 4:
        err = "Leech lattice vector must be of type 4"
        raise ValueError(err)
    vtype_beta = gen_leech2_subtype(v ^ BETA)
    if (vtype_beta >> 4) == 2:
        return reduce_type2_ortho(v ^ BETA, verbose)
    return reduce_type4_std(v, verbose)
     

#####################################################################
# Testing function  data for function reduce_type_4_std()


@pytest.mark.gen_xi
def test_reduce_type_4(ntests = 500, verbose = 0):
    """Test function ``reduce_type4`` """
    for n, v in enumerate(type4_testdata(ntests)):
        if verbose:
            print(" \nTest %d, v = %s, subtype  =%s" % 
                (n+1, hex(v), hex(gen_leech2_subtype(v)))
            )
        op = reduce_type4(v, verbose)
        w = gen_leech2_op_word(v, op, len(op)) 
        assert w & 0xffffff == 0x800000, hex(w)
        if gen_leech2_type(v ^ BETA) == 2:
            b = gen_leech2_op_word(BETA, op, len(op)) 
            assert b & 0xffffff == BETA, hex(b)
        a = np.zeros(6, dtype = np.uint32)
        l = gen_leech2_reduce_type4(v, a)
        if l < 0:
            err = "Error %s in function gen_leech2_reduce_type4"
            raise ValueError(err % hex(l & 0xffffffff))
        a = a[:l]
        assert list(a) == list(op), ((a), (op))  
            





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
import collections
import re
import warnings
from random import randint
import threading
import datetime
import pytest



from mmgroup.tests.groups.mgroup_n import MGroupNWord
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV


NJOBS = 5
GROUP_N_ONLY = 0

good_mm_orders = set(range(1,37)) | set(range(38,43)) | set(range(44,49))
good_mm_orders.update([50, 51, 52, 54, 55, 56, 57, 59, 60, 62, 66, 
    68, 69, 70, 71, 78, 84, 87, 88, 92, 93, 94, 95, 104, 105, 110, 119])

max_mm_order = max(good_mm_orders) 


p = 241
rep_mm = SparseMmV(p)
grp = MGroupNWord


def do_test_mm_order(v,  m, verbose = 0):
    v1, n = v.copy(), 0
    while n <= max_mm_order:
        v1, n = v1 * m, n+1
        if verbose: print("\r", n, end = "")
        if v1 == v:
             return n
    return None


v_tags = "ABCTTXXXYYYZZZ"

def rand_v():
    v = rep_mm()
    for s in v_tags:
        v +=  rep_mm(s, 'r')
    return v
    
    
def rand_m_monomial():
    m = grp()
    for s in "pxy":
        m *=  grp(s, 'r')
    if not GROUP_N_ONLY :
        m *= grp('t', randint(1,2))
    m *= grp('l', randint(1,2))   
    return m

def rand_m():
    m = grp()
    for i in range(4):
        m *=  rand_m_monomial()   
    return m

def do_test_order(display = True):
    v, m = rand_v(), rand_m()
    if display: 
        print("\rv =", v)
        print("\rm =", m)
    order = do_test_mm_order(v,  m, display)
    ok = order in good_mm_orders
    st = "ok" if ok else "error"
    if display: 
        print("\rorder is", order, ",", st)
    s =  "v = " + str(v) + "\nm = " + str(m) 
    s += "\norder = " + str(order) + ", " + st + "\n"
    return ok, s



@pytest.mark.space
@pytest.mark.slow
@pytest.mark.very_slow
@pytest.mark.extremely_slow
def test_orders(ntests = NJOBS):
    nerrors = 0
    for i in range(ntests):
        print("Test %d," % (i+1), datetime.datetime.now())
        ok, _ =  do_test_order()
        nerrors += not ok
    print("%d tests, %d errors, " % (ntests, nerrors),
            datetime.datetime.now())
    if nerrors:
        raise ValueError("Illegal order of monster group element found")









r"""We deal with an extraspecial group that maps to the Leech lattice modulo 2

Let :math:`Q_{x0}` be the group generated by elements :math:`x_d, x_\delta`
with :math:`d \in \mathcal{P}, \delta \in \mathcal{C}^*`. Here 
:math:`\mathcal{P}` is the Parker loop and :math:`\mathcal{C}^*` is the
Golay cocode as in :cite:`Con85`. We take the following relations in
:math:`Q_x` from :cite:`Seysen20`:


.. math::

    x_d x_e = x_{d \cdot e} x_{A(d,e)}, \,
    x_\delta x_\epsilon = x_{\delta \epsilon}, \,
    [x_d x_\delta] = x_{-1}^{\langle d, \delta \rangle}; \;
    d, e \in \mathcal{P}; \;
    \delta , \epsilon \in \mathcal{C}^* \, .

Here :math:`A(d,e)` is the associator between :math:`d` and :math:`e`,
i.e. the element of :math:`\mathcal{C}^*` corresponding to the vector 
:math:`d \cap e`; and :math:`[.,.]` is the commutator in :math:`Q_{x0}`.
Then :math:`Q_{x0}` is an extraspecal group of structure 
:math:`2_+^{1+24}`.

An element of the group :math:`Q_x` is modelled as an instance of
class |XLeech2|.

In that class we encode the element 
:math:`x_d \cdot x_\delta` of :math:`Q_{x0}` as an integer 
:math:`x`  as follows:

.. math::

       x = 2^{12} \cdot d \oplus (\delta \oplus \theta(d)) \, .

Here elements of the Parker loop and elements of the cocode 
are encoded as integers as in section :ref:`parker-loop-label`
and :ref:`golay-label`.
:math:`\theta` is the cocycle given in section 
:ref:`basis-golay-label`, and ':math:`\oplus`' means bitwise
addition modulo 2. Note that a  Parker loop element is 13 bits 
long (with the most significant bit denoting the sign) and that 
a cocode element is 12 bits long.

There is a natural homomorphism from :math:`Q_{x0}` to the 
Leech lattice :math:`\Lambda` modulo 2 with kernel 
:math:`\{x_1, x_{-1}\}`, as described in  :cite:`Con85`. This 
homomorphism maps the group operation in  :math:`Q_{x0}`
to vector addition in :math:`\Lambda / 2\Lambda`. With our
numbering in :math:`Q_{x0}` that homomorphism can be realized
by simply dropping the sign bit 25; then vector addition in
:math:`\Lambda / 2\Lambda` corresponds to the :math:`\oplus`
operation on such numbers. The natural quadratic form on
the vector  in :math:`\Lambda / 2\Lambda` with the number
:math:`2^{12} \cdot d + \delta`, :math:`0 \leq d, \delta < 2^{12}`
is the parity of the bit vector :math:`d \oplus \delta`.



"""




from functools import reduce
from operator import __xor__
from numbers import Integral, Number
from random import randint

import numpy as np

try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24


from mmgroup.mm import mm_aux_index_sparse_to_leech2
from mmgroup.mm import mm_aux_index_extern_to_sparse
from mmgroup.mm import mm_aux_index_leech2_to_sparse

from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.abstract_group import AbstractGroup
from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
from mmgroup.structures.parity import Parity
from mmgroup.structures.parse_atoms import ihex
from mmgroup.structures.mm_space_indices import tuple_to_sparse


from mmgroup.structures.gcode import GCode, GcVector
from mmgroup.structures.ploop import PLoop
from mmgroup.structures.cocode import Cocode
from mmgroup.structures.autpl import AutPL, AutPlGroup
from mmgroup.structures.construct_mm import add_to_embedded_classes

from mmgroup.generators import gen_leech2_mul
from mmgroup.generators import gen_leech2_pow
from mmgroup.generators import gen_leech2_scalprod
from mmgroup.generators import rand_get_seed, gen_leech2_type
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_op_word

ERR_RAND = "Illegal string for constricting type %s element" 

ERR_DIV4 = "%s object may be divided by 2 or 4 only"
ERR_DIV2 = "%s object may be divided by 2 only"


#######################################################################
# Import derived classed
#######################################################################

import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending, MM0
    from mmgroup.structures.mm0_group import MM0
    import_pending = False



###########################################################################
# Import functions from module ``mmgroup.mm_order`` on demand
###########################################################################


# Functions to be imported from module mmgroup.mm_order
check_mm_in_g_x0 = None

def import_mm_order_functions():
    """Import functions from module ``mmgroup.mm_order``.

    We import these functions from module ``mmgroup.mm_order``
    on demand. This avoids an infinite recursion of imports.
    """
    global check_mm_in_g_x0
    from mmgroup.structures.mm_order import check_mm_in_g_x0 



#######################################################################
# Auxiliary functions
#######################################################################


ERR_XL_TUPLE = "Cannot convert tuple to XLeech2 object"

ERR_XL_IN_Q = "Monster group element is not in subgroup Q_x0"

ERR_XL_TYPE = "Cannot convert '%s' object to XLeech2 object"

ERR_XL_RAND = "No random object in class XLeech2 found"

ERR_XL_SHORT = "XLeech2 object does not map to short Leech lattice vector"


def MM_to_Q_x0(g):
    if check_mm_in_g_x0 is None:
        import_mm_order_functions()
    g = MM0('a', g.mmdata)
    if check_mm_in_g_x0(g) is None:
        raise ValueError(ERR_XL_IN_Q)
    g.reduce()
    res = 0;
    for atom in g.mmdata:
        tag = (atom >> 28) & 0x0f
        if res == 0 and tag == 3:
            res = ((atom & 0x1fff) << 12) ^ mat24.ploop_theta(atom)
        elif tag == 1:
            res ^= atom & 0xfff
        elif tag:
            raise ValueError(ERR_XL_IN_Q)
    return res


def rand_xleech2_type(vtype):
    if vtype in [3,4]:
        for i in range(1000):
            v = randint(0, 0x1ffffff)
            if gen_leech2_type(v) == vtype:
                 return v
        raise ValueError(ERR_XL_RAND)
    if vtype == 0:
        return 0
    if vtype == 2:
        ve = randint(300, 98579)
        vs = mm_aux_index_extern_to_sparse(ve)
        sign = randint(0,1)
        return mm_aux_index_sparse_to_leech2(vs) ^ (sign << 24) 
    raise ValueError(ERR_XL_RAND)


def value_from_ploop(ploop=0, cocode = None, *args):
    c = Cocode(cocode).ord if cocode else 0
    if isinstance(ploop, Integral):
        d = ploop & 0x1ffffff
    elif isinstance(ploop, PLoop):
        d = ploop.value & 0x1fff
        d = (d << 12) ^ mat24.ploop_theta(d) 
    elif isinstance(ploop, XLeech2):
        d = ploop.value
    elif isinstance(ploop, Cocode):
        d = ploop.value & 0xfff
    elif isinstance(ploop, AbstractMMGroupWord):
        d = MM_to_Q_x0(ploop)
    elif isinstance(ploop, str):
        if len(ploop) == 1 and ploop in "BCTXES":
            d = 0
            a = tuple_to_sparse(0xff,  ploop, cocode, *args)
            if len(a) == 1:
                a0 = a[0]
                d = mm_aux_index_sparse_to_leech2(a0)
                a0 &= 0xff
                if a0 == 0xfe:
                    d ^= 0x1000000
                elif a0 != 1:
                    d = 0
            if d:
                return d
        if ploop == "r":
             if cocode is None:
                 return randint(0, 0x1ffffff)
             elif cocode in [0,2,3,4]:
                 return  rand_xleech2_type(cocode)
        raise ValueError(ERR_XL_TUPLE)            
    else:
        return TypeError(ERR_XL_TYPE % type(ploop))
    return d ^ c

        

#######################################################################
# Class XLeech2
#######################################################################





class XLeech2(AbstractGroupWord):
    r"""This class models an element of the group :math:`Q_{x0}`.

    The group :math:`Q_{x0}` is an extraspecial 2 group of structure 
    :math:`2^{1+24}`. 

    The group operation is written multiplicatively. 
        
    The :math:`2^{1+24}` group elements are numbered from
    ``0`` to ``0x1ffffff``. Elements ``0`` to ``0xffffff`` are
    considered positive. The element with number  ``0x1000000 ^ i`` 
    is the negative of the element with number  ``i``.

    An element is constructed as a product :math:`x_d x_\delta`,
    where :math:`d` is in the Parker loop and :math:`\delta` is in
    the Golay cocode.

    :param value:

      This parameter describes the value of an element of 
      the group :math:`Q_{x0}` as indicated in the table below. 

    :param cocode:

      This parameter describes an element  :math:`x_\delta` (with
      :math:`\delta` in the Golay cocode) 
      to be right multiplied to the element of :math:`Q_{x0}` obtained
      from the first parameter. 

    :return: An element of the group  :math:`Q_{x0}`.
    :rtype:  an instance of class |XLeech2|

    :raise:
        * ValueError if the input cannot converted to an element of
          the group :math:`Q_{x0}`.
        * TypeError the ype of an input is illegal.

    .. Caution::
       Although the group  :math:`Q_{x0}` has a natural embedding
       into the subgroup :math:`G_{x0}` of the monster group, we do 
       not consider an instance :math:`q` of class |XLeech2| as an 
       element of :math:`G_{x0}`. We consider :math:`q` as a (possibly
       negated) basis vector of a space on which the subgroup
       :math:`G_{x0}` of the monster acts monomially by right
       multiplication.



    Depending on its type parameter **value** is  interpreted as follows:

    .. table:: Legal types for constructor of class ``XLeech2``
      :widths: 20 80

      ===================== ================================================
      type of ``value``     Evaluates to
      ===================== ================================================
      ``int``               Here the code word with number ``value`` is
                            returned.  ``0 <= value < 0x2000000`` must hold.
                              
      class |XLeech2|       A deep copy of the object ``value``
                            is created. 

      class |GCode|         The corresponding Golay code element is
                            converted to a (positive) element of
                            :math:`Q_{x0}`. 

      class |PLoop|         The corresponding Parker loop element is
                            converted to an element of :math:`Q_{x0}`. 

      class |MM|            Then ``value`` in an element of the monster
                            group. If that element is in the subgroup
                            :math:`Q_{x0}` of the monster, then it is
                            converted to the coresponding instance of 
                            class |XLeech2|.

      ``'r'``               Create random element depending on the string,
                            see explanation below.

      A letter ``BCTXE``    If ``value`` is a single capital letter in
                            ``BCTXE`` then we create an element
                            corresponding to a unit vector 
                            in :math:`\rho_p` as described below.
      ===================== ================================================

    If value is of one of the type listed above then the following
    parameter is converted to an element of the Golay cocode (as in
    the constructor of class |Cocode| multiplied to the converted
    parmeter ``value`` from the right.

    Alternatively, parameter ``value`` may be a single letter. If
    ``value`` is one of the capital letters ``BCTXE`` then that letter
    and the following arguments are passed to the constructor of
    class |MMVector| in order to generate a (possible negated) basis
    vector of the representation :math:`\rho_p` of the monster.
    (Here the characteristic :math:`p` is ignored.) If that basis
    vector corresponds to an element of :math:`Q_{x0}` then we
    construct that element of :math:`Q_{x0}`; otherwise an error
    occurs.

    If parameter ``value`` is the letter ``'r'`` then we generate a
    random element of :math:`Q_{x0}`. If an integer ``i`` is given as a 
    second parameter ``'r'`` then we generate a random  element of 
    :math:`Q_{x0}` of type ``i``. Here the type of a vector 
    :math:`v \in \Lambda / 2 \Lambda` is the halved norm of the shortest
    vector in the set :math:`v + 2 \Lambda`. If ``i`` is specified
    then it must be equal to 0, 2, 3 or 4.
 
    **Standard operations**

    Let ``q`` be an instance of this class. The multiplication operator 
    ``*`` implements the group operation. Division by an element means
    multiplication by its inverse, and exponentiation means repeated 
    multiplication,  with ``q**(-1)`` the inverse of  ``q``,  as usual. 

    Multiplication with the integer ``1`` or ``-1`` means the 
    multiplication with the neutral element or with the central 
    involution :math:`x_{-1}`.
     
    The opration  ``&`` denotes the scalar product of the vectors
    in the Leech lattice modulo 2 obtained from an instance of this
    class, ignoring the sign.

    **Standard functions**
  
    ``abs(a)`` returns the element in the set ``{a, -a}`` which is
    positive.

    If an instance :math:`q` of class |XLeech2| maps to a vector of 
    type 2 in the Leech lattice modulo 2 then :math:`q` is also a unit 
    vector in the 196884-dimensional representation :math:`\rho` of 
    the monster group.

    One may use :math:`q` in the constructor of class |MM|
    (representing the monster group) for creating the corresponding
    element of the subgroup :math:`Q_{x0}` of the monster group.


    """
    __slots__ = "value",
 
    def __init__(self, ploop = 0, cocode = 0, *args):
        if import_pending:
            complete_import()
        self.value = value_from_ploop(ploop, cocode, *args)


    def __mul__(self, other):
        if isinstance(other, XLeech2):
            return XLeech2(gen_leech2_mul(self.value, other.value))
        elif isinstance(other, AbstractMMGroupWord):
            data = other.mmdata
            v = gen_leech2_op_word(self.value, data, len(data))
            return XLeech2(v)
        elif isinstance(other, Integral):
            if abs(other) == 1:
                return XLeech2(self.value ^ ((other & 2) << 23))
            elif other == 0:
                 return XLeech2(0)
            return NotImplemented
        else:           
            return NotImplemented

    def __imul__(self, other):
        self.value = self.__mul__(other).value
        return self

    def __rmul__(self, other):
        if isinstance(other, XLeech2):
            return XLeech2(gen_leech2_mul(other.value, self.value))
        elif isinstance(other, Integral):
            if abs(other) == 1:
                return XLeech2(self.value ^ ((other & 2) << 23))
            elif other == 0:
                 return XLeech2(0)
            return NotImplemented
        else:           
            return NotImplemented

    def __abs__(self):
        return  XLeech2(self.value & 0xffffff)

    def __pos__(self):
        return  self

    def __neg__(self):
        return  XLeech2(self.value ^ 0x1000000)

    def __pow__(self, other):
        if isinstance(other, Integral):            
            return XLeech2(gen_leech2_pow(self.value, other & 3))
        elif isinstance(other, XLeech2):
            ov = other.value
            v = self.value
            w = gen_leech2_mul(gen_leech2_pow(ov, 3), v)
            w = gen_leech2_mul(w, ov)
            return XLeech2(w)
        else:
            return NotImplemented
        
    def __truediv__(self, other):
        if isinstance(other, XLeech2):
            v = gen_leech2_pow(other.value, 3)
            return XLeech2(gen_leech2_mul(self.value, other.value))
        elif isinstance(other, Integral):
            if abs(other) == 1:
                v = (other & 2) << 23
            else:
                return NotImplemented  
        else:           
            return NotImplemented
        return XLeech2(gen_leech2_mul(self.value, v))

    def __itruediv__(self, other):
        self.value = self.__itruediv__(other).value
        return self


                             
    def __rtruediv__(self, other):
        v = gen_leech2_pow(other.value, 3)
        if isinstance(other, XLeech2):
            return XLeech2(gen_leech2_mul(other.value, v))
        elif isinstance(other, Integral):
            if abs(other) == 1:
                return XLeech2(v ^ ((other & 2) << 23))
            elif other == 0:
                 return XLeech2(0)
            return NotImplemented
        else:           
            return NotImplemented
 
       
    def __and__(self, other):
        if isinstance(other, XLeech2):
            ov = other
        elif isinstance(other, (GCode, PLoop)):
            d = other.value & 0xfff
            ov = (d << 12) ^ mat24.ploop_theta(d) 
        elif isinstance(ploop, Cocode):
            ov = ploop.value & 0xfff
        else:
            return NotImplemented
        return Parity(gen_leech2_scalprod(self.value, ov))


    __rand__ = __and__

    def __eq__(self, other):
        return (isinstance(other, XLeech2)  
            and (self.value ^ other.value) & 0x1ffffff == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self
        
    def __neg__(self):
        return  XLeech2(self.value ^ 0x1000000)


    @property
    def ord(self):
        r"""Return the number of the element of :math:`Q_{x0}` .

        We have ``0 <= i < 0x2000000`` for the returned number ``i``.
        """
        return self.value & 0x1ffffff


    @property
    def sign(self):
        r"""Return the sign of the element :math:`Q_{x0}`.

        This is ``1`` for a positive and ``-1`` for a negative element.
        """
        return 1 - ((self.value >> 23) & 2)

    def isplit(self):
        r"""Split element into a product :math:`x_d \cdot x_\delta`

        The method returns a pair  :math:`(d,  \delta)` such that
        :math:`x_d \cdot x_\delta` is equal to the given element
        of  :math:`Q_{x0}`.

        Here :math:`(d,  \delta)` is returned as a pair of integers.
        """
        v = self.value
        x = (v >> 12) & 0x1fff
        d = (mat24.ploop_theta(v >> 12) ^ v) & 0xfff
        return x, d


    def split(self):
        r"""Split element into a product :math:`x_d \cdot x_\delta`

        The method returns a pair  :math:`(d,  \delta)` such that
        :math:`x_d \cdot x_\delta` is equal to the given element
        of  :math:`Q_{x0}`.

        The element :math:`d` of the Parker loop is an instance of
        class |PLoop|. The element  :math:`\delta` of the Golay cocode
        is an instance of class |Cocode|.
        """
        x, d = self.isplit()
        return PLoop(x), Cocode(d)
 
    @property
    def mmdata(self):
        """Return internal representation of corresponding monster element.

        That internal representation is returned as a numpy array 
        of 32-bit integers.
        """
        v = self.value
        x = (v >> 12) & 0x1fff
        d = (mat24.ploop_theta(v >> 12) ^ v) & 0xfff
        return np.array([0x30000000 + x, 0x10000000 + d], 
            dtype = np.uint32)
     
    def str(self):
        v = self.value
        x = (v >> 12) & 0x1fff
        d = (mat24.ploop_theta(v >> 12) ^ v) & 0xfff
        return "XL2<x_%s*d_%s>" % (ihex(x, 3), ihex(d, 3))
    __repr__  = str


    
    @property
    def type(self):
        r"""Return the type of the element of :math:`Q_{x0}`.

        This is equal to the type of the corresponding vector
        :math:`v` in the Leech lattice :math:`\Lambda` modulo 2. 
        The type of a vector :math:`v \in \Lambda / 2 \Lambda` 
        is the halved norm of the shortest  vector in the set 
        :math:`v + 2 \Lambda`. That type is equal to 0, 2, 3 or 4.
        """
        return gen_leech2_type(self.value)

    @property
    def xsubtype(self):
        r"""Return subtype of the element of :math:`Q_{x0}`

        The function returns the integer ``16 * type + subtype``,
        where the pair ``(type, subtype)`` is as in property
        ``subtype``. See section :ref:`computation-leech2` in
        the **guide** for background.
        """
        return gen_leech2_subtype(self.value)

    @property
    def subtype(self):
        r"""Return pair ``(type, subtype)`` of the element of :math:`Q_{x0}`

        Here ``type`` is as in property ``type``, and subtype is
        a one-digit decimal integer describing the subtype of
        a vector in the Leech lattice modulo 2, as explained in
        the **guide** in section :ref:`computation-leech2`.
        """
        t =  gen_leech2_subtype(self.value)
        return t >> 4, t & 15


    def as_Q_x0_atom(self):
        return self.value & 0x1ffffff

    def vector_tuple(self):
        r"""Return unit vector in representation of the monster as tuple

        If the element of :math:`Q_{x0}` corresponds to a vector of
        type 2 in the Leech lattice then it also corresponds to a
        unit vector ``v`` in the representation  :math:`\rho` 
        of :math:`\mathbb{M}`; otherwise the function raises
        ValueError. 
        
        The function returns a  tuple ``(sign, tag, i0, i1)``, with  
        ``sign = +1 or -1``. Then the triple ``(tag, i0, i1)`` 
        corresponds to a basis vector ``u`` in  :math:`\rho` as 
        described in class |MMVector|, and we have ``v = sign * u``.         
        """
        sign = 1 - ((self.value >> 23) & 2)
        sp = mm_aux_index_leech2_to_sparse(self.value & 0xffffff)
        if sp == 0:
            raise ValueError(ERR_XL_SHORT)
        tag = "?ABCTXZY"[(sp >> 25) & 7]
        i0, i1 = (sp >> 14) & 0x7ff, (sp >> 8) & 0x3f
        return sign, tag, i0, i1

    def octad_number(self):
        r"""Return number of octad.

        Let :math:`x_d \cdot x_\delta` be equal to the given element
        of  :math:`Q_{x0}`. If :math:`d` corresponds to a (possibly
        complemented) octad in the Golay code then the function 
        returns the number of that octad as a nonnegative integer 
        less than 759. Otherwise the function raises ``ValueError``.

        Here the octads are numbered as in class |GCode|.
        """
        return mat24.gcode_to_octad(self.value >> 12, 0)


add_to_embedded_classes(XLeech2)







from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import warnings
import copy
import numbers
import re
import numpy as np
from numbers import Integral


#from mmgroup.mm import mm_aux_index_sparse_to_leech2
from mmgroup import MM0
from mmgroup.mm_space import MMSpace, MMV
from mmgroup.mm import INT_BITS, PROTECT_OVERFLOW

from mmgroup.generators import gen_leech2to3_short
from mmgroup.clifford12 import xsp2co1_elem_to_qs, xsp2co1_qs_to_elem_i 
from mmgroup.clifford12 import error_string, chk_qstate12
from mmgroup.clifford12 import xsp2co1_neg_elem, xsp2co1_elem_row_mod3
from mmgroup.clifford12 import xsp2co1_mul_elem, xsp2co1_reduce_elem
from mmgroup.structures.xsp2_co1 import get_error_pool
from mmgroup.structures.xsp2_co1 import Xsp2_Co1, str_xsp2_co1
from mmgroup.structures.qs_matrix import QStateMatrix
from mmgroup.structures.abstract_rep_space import AbstractRepVector
from mmgroup.structures.abstract_rep_space import AbstractRepSpace
#from mmgroup.structures.mm_space_indices import sparse_to_tuples




MMSpace3 =  MMV(3)
MMGroup3 = MM0

######################################################################
# Creating a leech lattice vector mod 3 from a tuple
######################################################################

ERR_LEECH = "Tuple does not describe a short Leech lattice vector %s"


def tuple_to_leech_mod3(tag, i0 = -1, i1 = -1):
    leech_mod2 = MMSpace.index_to_short_mod2(tag, i0, i1)
    res = gen_leech2to3_short(leech_mod2)
    if (res == 0):
        err = get_error_pool(15)
        if len(err):
            print("tuple_to_leech_mod3 input:", hex(leech_mod2))
            print("Leech", MMSpace.index_to_short(tag, i0, i1))
        raise ValueError(ERR_LEECH % "mod 3")
    return res


CODE_MOD3 = {0:0, 1:1, 2:0x1000001}
def array_to_leech_mod3(a):
    return sum(CODE_MOD3[x % 3] << i for i, x in enumerate(a[:24]))


######################################################################
# Creating a vector in rep 4096x as an instance of class QStateMatrix
######################################################################

ERR_QSTATE_SHAPE = "QStateMatrix object must have shape (0,12) here"

def obj_to_qstate(obj):
    if isinstance(obj, QStateMatrix):
        if obj.shape == (0,12):
            return obj.copy().extend_zero(12,0).reshape((12,12))
        if obj.shape == (12,12):
            return obj
        raise ValueError(ERR_QSTATE_SHAPE)
    if isinstance(obj, Integral):
        res = QStateMatrix(0, 12,  obj & 0xfff)
        if (obj & 0x1000):
            res = - res 
        return res.extend_zero(0,12).reshape((12,12))
    ERR = "Illegal type for creating an Xsp2_Co1_Vector"
    raise TypeError(ERR)  


######################################################################
# Modelling a vector
######################################################################

class Xsp2_Co1_Vector(AbstractRepVector):
    def __init__(self, space):
        __slots__ = 'space', '_data', 'is_zero'
        self.space = space
        self.is_zero = True
        self._data = np.zeros(26, dtype = np.uint64)

    @property
    def short3(self):
        return int(self._data[0])

    @property
    def short3_list(self):
        x = self._data[0]
        return  [(((x >> i) & 1) + ((x >> (23 + i)) & 2)) % 3 
            for i in range(24)]

    @property
    def qs(self):
        return QStateMatrix(xsp2co1_elem_to_qs(self._data))

    def _set_vector(self, rep_4096, rep_leech):
        qs = obj_to_qstate(rep_4096)
        if isinstance(rep_leech, Integral): 
            x = rep_leech
        else:
            tag = rep_leech[0]
            if isinstance(tag, str):
                x = tuple_to_leech_mod3(*rep_leech)
            else:
                x = array_to_leech_mod3(rep_leech)
        self._data[:] = xsp2co1_qs_to_elem_i(qs, x)
        self.is_zero = False
       
    def _set_zero(self):
        self.is_zero = True

    def check(self):
        """Check if the vector is correct

        Raise ValueError if the vector is errorneous.
        """
        self.space.check(self)
        
    def as_tuples(self):
        return self.space.as_tuples(self)   
        
    def dump(self):
        self.space.dump_vector(self)    

    def as_mmspace_vector(self):
        return self.space.as_mmspace_vector(self) 



######################################################################
# class SparseMmSpace
######################################################################




class Xsp2_Co1_Space(AbstractRepSpace):
    """Models the sparse representation 198884x of the monster group. 

    YET TO BE DOCUMENTED !!!

    """

    vector_type = Xsp2_Co1_Vector
    mmspace = MMSpace3

    def __init__(self):
        """Create a 196884-dimensional representation of the monster

        All calculations are done modulo the odd number p
        """
        super(Xsp2_Co1_Space, self).__init__(3, Xsp2_Co1)


    ### Standard methods that must be overwritten ####################

    def zero(self):
        return self.vector_type(self)

    def unit(self, rep_4096, rep_leech):
        v = self.vector_type(self)
        v._set_vector(rep_4096, rep_leech)
        return v

    def iadd(self, v1, v2):
        if v1.is_zero:
            return v2
        if v2.is_zero:
            return v1
        ERR = "Vector addition not supported in this space"
        raise TypeError(ERR)

    def imul_scalar(self, v1, a):
        a = int(a % 3)
        v1.is_zero =  v1.is_zero or a == 0
        if a == 2 and not v1.is_zero:
            xsp2co1_neg_elem(v1._data)
        return v1

    def imul_group_word(self, v1, g):
        if v1.is_zero:
            return v1
        chk_qstate12(xsp2co1_mul_elem(v1._data, g._data, v1._data))
        return v1

    def vector_get_item(self, v1, index):
        return self.as_mmspace_vector(v1)[index]

    def vector_set_item(self, v1, index, value) :
        ERR = "Item assigment not supported in this vector space"
        raise TypeError(ERR)

    def equal_vectors(self, v1, v2):
        if v1.is_zero != v2.is_zero:
            return False
        if v1.is_zero:
            return True
        self.reduce(v1)
        self.reduce(v2)
        return (v1._data == v2._data).all()
 

    ### Standard methods that must be overwritten ####################

    def copy_vector(self, v1):
        """Return deep copy of group element v1"""
        assert v1.space == self
        v2 = self.zero()
        v2._data[:] = v1._data
        v2.is_zero = v1.is_zero
        return v2

    def reduce(self, v):
        if not v.is_zero:
            chk_qstate12(xsp2co1_reduce_elem(v._data))
        return v


    def str_vector(self, v):
        """Convert vector v to a string

        """
        if v.is_zero:
            return "Xsp2 vector 0"
        return "Xsp2 vector " + str_xsp2_co1(v._data, factor=64, t=True)

    ### Standard methods that need not be overwritten #################

    def parse(self, s):
        ERR = "Cannot convert a strig to a  Xsp2_Co1_Vector"
        raise TypeError(ERR)

    ### Extra methods ##################################################

    def as_mmspace_vector(self, v1):
        vm = self.mmspace(0)
        assert v1.qs.shape == (12,12) 
        if INT_BITS == 64:
            START = 116416//32
            assert len(vm.data) >=  START + 4096
            chk_qstate12(xsp2co1_elem_row_mod3(v1._data, 0, 
                 vm.data[START:]))
            return vm
        ERR = "Function not supported on %d-bit systems"
        raise NotImplementedError(ERR % INT_BITS)

    def as_sparse(self, v1):
        """Yet to be documented!!

        """
        vm = self.as_mmspace_vector(v1)
        return vm.space().as_sparse(vm)

    def as_tuples(self, v1):
        """Return vector in tuple representation

        The function returns a list of tuples ``(factor, tag, i0, i1)``.
        Here ``(tag, i0, i1)`` describes a basis vector and ``value``
        is the coordinate of that vector.

        Entries with coordinate zero are dropped.
        """
        vm = self.as_mmspace_vector(v1)
        return vm.as_tuples()
        
    def dump_vector(self, v1):
        a = np.zeros(4096, dtype = np.uint64)    
        chk_qstate12(xsp2co1_elem_row_mod3(v1._data, 0, a))
        or_sum = 0
        print("Xsp2_Co1 vector:")
        for i, x in enumerate(a):
            x = int(x) & 0xffffffffffff
            if (x): 
                print("0x%04x: 0x%012x" % (i, x))
            or_sum |= x
        if not or_sum:
            print("Vector is equal to 0")

    def check(self, v1):
        pass  # Yet nothing do do here!!
        return v1

Space_ZY =  Xsp2_Co1_Space()        
